"""Resolwe models utils."""
import copy
import json
import os
import re

import jsonschema

from django.contrib.staticfiles import finders
from django.core.exceptions import ValidationError

from resolwe.flow.utils import dict_dot, iterate_dict, iterate_fields, iterate_schema


class DirtyError(ValidationError):
    """Error raised when required fields missing."""


def validation_schema(name):
    """Return json schema for json validation."""
    schemas = {
        "processor": "processSchema.json",
        "descriptor": "descriptorSchema.json",
        "field": "fieldSchema.json",
        "type": "typeSchema.json",
    }

    if name not in schemas:
        raise ValueError()

    field_schema_file = finders.find("flow/{}".format(schemas["field"]), all=True)[0]
    with open(field_schema_file, "r") as fn:
        field_schema = fn.read()

    if name == "field":
        return json.loads(field_schema.replace("{{PARENT}}", ""))

    schema_file = finders.find("flow/{}".format(schemas[name]), all=True)[0]
    with open(schema_file, "r") as fn:
        schema = fn.read()

    return json.loads(
        schema.replace("{{FIELD}}", field_schema).replace("{{PARENT}}", "/field")
    )


TYPE_SCHEMA = validation_schema("type")


def validate_schema(
    instance, schema, test_required=True, data_location=None, skip_missing_data=False
):
    """Check if DictField values are consistent with our data types.

    Perform basic JSON schema validation and our custom validations:

      * check that required fields are given (if `test_required` is set
        to ``True``)
      * check if ``basic:file:`` and ``list:basic:file`` fields match
        regex given in schema (only if ``validate_regex`` is defined in
        schema for coresponding fields) and exists (only if
        ``data_location`` is given)
      * check if directories referenced in ``basic:dir:`` and
        ``list:basic:dir``fields exist (only if ``data_location`` is
        given)
      * check that referenced ``Data`` objects (in ``data:<data_type>``
        and  ``list:data:<data_type>`` fields) exists and are of type
        ``<data_type>``
      * check that referenced ``Storage`` objects (in ``basic:json``
        fields) exists

    :param list instance: Instance to be validated
    :param list schema: Schema for validation
    :param bool test_required: Flag for testing if all required fields
        are present. It is usefule if validation is run before ``Data``
        object is finished and there are some field stil missing
        (default: ``False``)
    :param :class:`~resolwe.flow.models.data.DataLocation` data_location:
        data location used for checking if files and directories exist
        (default: ``None``)
    :param bool skip_missing_data: Don't raise an error if referenced
        ``Data`` object does not exist
    :rtype: None
    :raises ValidationError: if ``instance`` doesn't match schema
        defined in ``schema``

    """
    from .storage import Storage  # Prevent circular import.

    path_prefix = None
    if data_location:
        path_prefix = data_location.get_path()

    def validate_refs(field):
        """Validate reference paths."""
        for ref_filename in field.get("refs", []):
            ref_path = os.path.join(path_prefix, ref_filename)
            if not os.path.exists(ref_path):
                raise ValidationError(
                    "Path referenced in `refs` ({}) does not exist.".format(ref_path)
                )
            if not (os.path.isfile(ref_path) or os.path.isdir(ref_path)):
                raise ValidationError(
                    "Path referenced in `refs` ({}) is neither a file or directory.".format(
                        ref_path
                    )
                )

    def validate_file(field, regex):
        """Validate file name (and check that it exists)."""
        filename = field["file"]

        if regex and not re.search(regex, filename):
            raise ValidationError(
                "File name {} does not match regex {}".format(filename, regex)
            )

        if path_prefix:
            path = os.path.join(path_prefix, filename)
            if not os.path.exists(path):
                raise ValidationError(
                    "Referenced path ({}) does not exist.".format(path)
                )
            if not os.path.isfile(path):
                raise ValidationError(
                    "Referenced path ({}) is not a file.".format(path)
                )

            validate_refs(field)

    def validate_dir(field):
        """Check that dirs and referenced files exists."""
        dirname = field["dir"]

        if path_prefix:
            path = os.path.join(path_prefix, dirname)
            if not os.path.exists(path):
                raise ValidationError(
                    "Referenced path ({}) does not exist.".format(path)
                )
            if not os.path.isdir(path):
                raise ValidationError(
                    "Referenced path ({}) is not a directory.".format(path)
                )

            validate_refs(field)

    def validate_data(data_pk, type_):
        """Check that `Data` objects exist and is of right type."""
        from .data import Data  # prevent circular import

        data_qs = Data.objects.filter(pk=data_pk).values("process__type")
        if not data_qs.exists():
            if skip_missing_data:
                return

            raise ValidationError(
                "Referenced `Data` object does not exist (id:{})".format(data_pk)
            )
        data = data_qs.first()
        if not data["process__type"].startswith(type_):
            raise ValidationError(
                "Data object of type `{}` is required, but type `{}` is given. "
                "(id:{})".format(type_, data["process__type"], data_pk)
            )

    def validate_range(value, interval, name):
        """Check that given value is inside the specified range."""
        if not interval:
            return

        if value < interval[0] or value > interval[1]:
            raise ValidationError(
                "Value of field '{}' is out of range. It should be between {} and {}.".format(
                    name, interval[0], interval[1]
                )
            )

    is_dirty = False
    dirty_fields = []
    for _schema, _fields, _ in iterate_schema(instance, schema):
        name = _schema["name"]
        is_required = _schema.get("required", True)

        if test_required and is_required and name not in _fields:
            is_dirty = True
            dirty_fields.append(name)

        if name in _fields:
            field = _fields[name]
            type_ = _schema.get("type", "")

            # Treat None as if the field is missing.
            if not is_required and field is None:
                continue

            try:
                jsonschema.validate([{"type": type_, "value": field}], TYPE_SCHEMA)
            except jsonschema.exceptions.ValidationError as ex:
                raise ValidationError(ex.message)

            choices = [choice["value"] for choice in _schema.get("choices", [])]
            allow_custom_choice = _schema.get("allow_custom_choice", False)
            if choices and not allow_custom_choice and field not in choices:
                raise ValidationError(
                    "Value of field '{}' must match one of predefined choices. "
                    "Current value: {}".format(name, field)
                )

            if type_ == "basic:file:":
                validate_file(field, _schema.get("validate_regex"))

            elif type_ == "list:basic:file:":
                for obj in field:
                    validate_file(obj, _schema.get("validate_regex"))

            elif type_ == "basic:dir:":
                validate_dir(field)

            elif type_ == "list:basic:dir:":
                for obj in field:
                    validate_dir(obj)

            elif (
                type_ == "basic:json:" and not Storage.objects.filter(pk=field).exists()
            ):
                raise ValidationError(
                    "Referenced `Storage` object does not exist (id:{})".format(field)
                )

            elif type_.startswith("data:"):
                validate_data(field, type_)

            elif type_.startswith("list:data:"):
                for data_id in field:
                    validate_data(data_id, type_[5:])  # remove `list:` from type

            elif type_ == "basic:integer:" or type_ == "basic:decimal:":
                validate_range(field, _schema.get("range"), name)

            elif type_ == "list:basic:integer:" or type_ == "list:basic:decimal:":
                for obj in field:
                    validate_range(obj, _schema.get("range"), name)

    try:
        # Check that schema definitions exist for all fields
        for _, _ in iterate_fields(instance, schema):
            pass
    except KeyError as ex:
        raise ValidationError(str(ex))

    if is_dirty:
        dirty_fields = ['"{}"'.format(field) for field in dirty_fields]
        raise DirtyError(
            "Required fields {} not given.".format(", ".join(dirty_fields))
        )


def _hydrate_values(output, output_schema, data):
    """Hydrate basic:file and basic:json values.

    Find fields with basic:file type and assign a full path to the file.
    Find fields with basic:json type and assign a JSON object from storage.

    """

    def hydrate_path(file_name):
        """Hydrate file paths."""
        from resolwe.flow.managers import manager

        class HydratedPath(str):
            """String wrapper, which also stores the original filename."""

            __slots__ = ("data_id", "file_name")

            def __new__(cls, value=""):
                """Initialize hydrated path."""
                hydrated = str.__new__(cls, value)
                hydrated.data_id = data.id
                hydrated.file_name = file_name
                return hydrated

        return HydratedPath(manager.get_executor().resolve_data_path(data, file_name))

    def hydrate_storage(storage_id):
        """Hydrate storage fields."""
        from .storage import LazyStorageJSON  # Prevent circular import.

        return LazyStorageJSON(pk=storage_id)

    for field_schema, fields in iterate_fields(output, output_schema):
        name = field_schema["name"]
        value = fields[name]
        if "type" in field_schema:
            if field_schema["type"].startswith("basic:file:"):
                value["file"] = hydrate_path(value["file"])
                value["refs"] = [hydrate_path(ref) for ref in value.get("refs", [])]

            elif field_schema["type"].startswith("list:basic:file:"):
                for obj in value:
                    obj["file"] = hydrate_path(obj["file"])
                    obj["refs"] = [hydrate_path(ref) for ref in obj.get("refs", [])]

            if field_schema["type"].startswith("basic:dir:"):
                value["dir"] = hydrate_path(value["dir"])
                value["refs"] = [hydrate_path(ref) for ref in value.get("refs", [])]

            elif field_schema["type"].startswith("list:basic:dir:"):
                for obj in value:
                    obj["dir"] = hydrate_path(obj["dir"])
                    obj["refs"] = [hydrate_path(ref) for ref in obj.get("refs", [])]

            elif field_schema["type"].startswith("basic:json:"):
                fields[name] = hydrate_storage(value)

            elif field_schema["type"].startswith("list:basic:json:"):
                fields[name] = [hydrate_storage(storage_id) for storage_id in value]


def hydrate_input_references(input_, input_schema, hydrate_values=True):
    """Hydrate ``input_`` with linked data.

    Find fields with complex data:<...> types in ``input_``.
    Assign an output of corresponding data object to those fields.

    """
    from .data import Data  # prevent circular import

    for field_schema, fields in iterate_fields(input_, input_schema):
        name = field_schema["name"]
        value = fields[name]
        if "type" in field_schema:
            if field_schema["type"].startswith("data:"):
                if value is None:
                    continue

                try:
                    data = Data.objects.get(id=value)
                except Data.DoesNotExist:
                    fields[name] = {}
                    continue

                output = copy.deepcopy(data.output)
                hydrate_input_references(output, data.process.output_schema)
                if hydrate_values:
                    _hydrate_values(output, data.process.output_schema, data)
                output["__id"] = data.id
                output["__type"] = data.process.type
                output["__descriptor"] = data.descriptor
                output["__name"] = getattr(data, "name", None)
                output["__entity_name"] = getattr(data.entity, "name", None)
                output["__output_schema"] = data.process.output_schema

                fields[name] = output

            elif field_schema["type"].startswith("list:data:"):
                outputs = []
                for val in value:
                    if val is None:
                        continue

                    try:
                        data = Data.objects.get(id=val)
                    except Data.DoesNotExist:
                        outputs.append({})
                        continue

                    output = copy.deepcopy(data.output)
                    hydrate_input_references(output, data.process.output_schema)
                    if hydrate_values:
                        _hydrate_values(output, data.process.output_schema, data)

                    output["__id"] = data.id
                    output["__type"] = data.process.type
                    output["__descriptor"] = data.descriptor
                    output["__name"] = getattr(data, "name", None)
                    output["__entity_name"] = getattr(data.entity, "name", None)
                    output["__output_schema"] = data.process.output_schema

                    outputs.append(output)

                fields[name] = outputs


def hydrate_input_uploads(input_, input_schema, hydrate_values=True):
    """Hydrate input basic:upload types with upload location.

    Find basic:upload fields in input.
    Add the upload location for relative paths.

    """
    from resolwe.flow.managers import manager

    files = []
    for field_schema, fields in iterate_fields(input_, input_schema):
        name = field_schema["name"]
        value = fields[name]
        if "type" in field_schema:
            if field_schema["type"] == "basic:file:":
                files.append(value)

            elif field_schema["type"] == "list:basic:file:":
                files.extend(value)

    urlregex = re.compile(
        r"^(https?|ftp)://[-A-Za-z0-9\+&@#/%?=~_|!:,.;]*[-A-Za-z0-9\+&@#/%=~_|]"
    )
    for value in files:
        if "file_temp" in value:
            if isinstance(value["file_temp"], str):
                # If file_temp not url, hydrate path.
                if not urlregex.search(value["file_temp"]):
                    value["file_temp"] = manager.get_executor().resolve_upload_path(
                        value["file_temp"]
                    )
            else:
                # Something very strange happened.
                value["file_temp"] = "Invalid value for file_temp in DB"


def hydrate_size(data, force=False):
    """Add file and dir sizes.

    Add sizes to ``basic:file:``, ``list:basic:file``, ``basic:dir:``
    and ``list:basic:dir:`` fields.

    ``force`` parameter is used to recompute file sizes also on objects
    that already have these values, e.g. in migrations.
    """
    from .data import Data  # prevent circular import

    def get_dir_size(path):
        """Get directory size."""
        total_size = 0
        for dirpath, _, filenames in os.walk(path):
            for file_name in filenames:
                file_path = os.path.join(dirpath, file_name)
                if not os.path.isfile(
                    file_path
                ):  # Skip all "not normal" files (links, ...)
                    continue
                total_size += os.path.getsize(file_path)
        return total_size

    def get_refs_size(obj, obj_path):
        """Calculate size of all references of ``obj``.

        :param dict obj: Data object's output field (of type file/dir).
        :param str obj_path: Path to ``obj``.
        """
        total_size = 0
        for ref in obj.get("refs", []):
            ref_path = data.location.get_path(filename=ref)
            if ref_path in obj_path:
                # It is a common case that ``obj['file']`` is also contained in
                # one of obj['ref']. In that case, we need to make sure that it's
                # size is not counted twice:
                continue
            if os.path.isfile(ref_path):
                total_size += os.path.getsize(ref_path)
            elif os.path.isdir(ref_path):
                total_size += get_dir_size(ref_path)

        return total_size

    def add_file_size(obj):
        """Add file size to the basic:file field."""
        if (
            data.status in [Data.STATUS_DONE, Data.STATUS_ERROR]
            and "size" in obj
            and not force
        ):
            return

        path = data.location.get_path(filename=obj["file"])
        if not os.path.isfile(path):
            raise ValidationError("Referenced file does not exist ({})".format(path))

        obj["size"] = os.path.getsize(path)
        obj["total_size"] = obj["size"] + get_refs_size(obj, path)

    def add_dir_size(obj):
        """Add directory size to the basic:dir field."""
        if (
            data.status in [Data.STATUS_DONE, Data.STATUS_ERROR]
            and "size" in obj
            and not force
        ):
            return

        path = data.location.get_path(filename=obj["dir"])
        if not os.path.isdir(path):
            raise ValidationError("Referenced dir does not exist ({})".format(path))

        obj["size"] = get_dir_size(path)
        obj["total_size"] = obj["size"] + get_refs_size(obj, path)

    data_size = 0
    for field_schema, fields in iterate_fields(data.output, data.process.output_schema):
        name = field_schema["name"]
        value = fields[name]
        if "type" in field_schema:
            if field_schema["type"].startswith("basic:file:"):
                add_file_size(value)
                data_size += value.get("total_size", 0)
            elif field_schema["type"].startswith("list:basic:file:"):
                for obj in value:
                    add_file_size(obj)
                    data_size += obj.get("total_size", 0)
            elif field_schema["type"].startswith("basic:dir:"):
                add_dir_size(value)
                data_size += value.get("total_size", 0)
            elif field_schema["type"].startswith("list:basic:dir:"):
                for obj in value:
                    add_dir_size(obj)
                    data_size += obj.get("total_size", 0)

    data.size = data_size


def render_descriptor(data):
    """Render data descriptor.

    The rendering is based on descriptor schema and input context.

    :param data: data instance
    :type data: :class:`resolwe.flow.models.Data` or :class:`dict`

    """
    if not data.descriptor_schema:
        return

    # Set default values
    for field_schema, field, path in iterate_schema(
        data.descriptor, data.descriptor_schema.schema, "descriptor"
    ):
        if "default" in field_schema and field_schema["name"] not in field:
            dict_dot(data, path, field_schema["default"])


def render_template(process, template_string, context):
    """Render template using the specified expression engine."""
    from resolwe.flow.managers import manager

    # Get the appropriate expression engine. If none is defined, do not evaluate
    # any expressions.
    expression_engine = process.requirements.get("expression-engine", None)
    if not expression_engine:
        return template_string

    return manager.get_expression_engine(expression_engine).evaluate_block(
        template_string, context
    )


def json_path_components(path):
    """Convert JSON path to individual path components.

    :param path: JSON path, which can be either an iterable of path
        components or a dot-separated string
    :return: A list of path components
    """
    if isinstance(path, str):
        path = path.split(".")

    return list(path)


def validate_process_subtype(supertype_name, supertype, subtype_name, subtype):
    """Perform process subtype validation.

    :param supertype_name: Supertype name
    :param supertype: Supertype schema
    :param subtype_name: Subtype name
    :param subtype: Subtype schema
    :return: A list of validation error strings
    """
    errors = []
    for item in supertype:
        # Ensure that the item exists in subtype and has the same schema.
        for subitem in subtype:
            if item["name"] != subitem["name"]:
                continue

            for key in set(item.keys()) | set(subitem.keys()):
                if key in ("label", "description"):
                    # Label and description can differ.
                    continue
                elif key == "required":
                    # A non-required item can be made required in subtype, but not the
                    # other way around.
                    item_required = item.get("required", True)
                    subitem_required = subitem.get("required", False)

                    if item_required and not subitem_required:
                        errors.append(
                            "Field '{}' is marked as required in '{}' and optional in '{}'.".format(
                                item["name"], supertype_name, subtype_name,
                            )
                        )
                elif item.get(key, None) != subitem.get(key, None):
                    errors.append(
                        "Schema for field '{}' in type '{}' does not match supertype '{}'.".format(
                            item["name"], subtype_name, supertype_name
                        )
                    )

            break
        else:
            errors.append(
                "Schema for type '{}' is missing supertype '{}' field '{}'.".format(
                    subtype_name, supertype_name, item["name"]
                )
            )

    return errors


def validate_process_types(queryset=None):
    """Perform process type validation.

    :param queryset: Optional process queryset to validate
    :return: A list of validation error strings
    """
    if not queryset:
        from .process import Process

        queryset = Process.objects.all()

    processes = {}
    for process in queryset:
        dict_dot(
            processes,
            process.type.replace(":", ".") + "__schema__",
            process.output_schema,
        )

    errors = []
    for path, key, value in iterate_dict(
        processes, exclude=lambda key, value: key == "__schema__"
    ):
        if "__schema__" not in value:
            continue

        # Validate with any parent types.
        for length in range(len(path), 0, -1):
            parent_type = ".".join(path[:length] + ["__schema__"])
            try:
                parent_schema = dict_dot(processes, parent_type)
            except KeyError:
                continue

            errors += validate_process_subtype(
                supertype_name=":".join(path[:length]),
                supertype=parent_schema,
                subtype_name=":".join(path + [key]),
                subtype=value["__schema__"],
            )

    return errors


def fill_with_defaults(process_input, input_schema):
    """Fill empty optional fields in input with default values."""
    for field_schema, fields, path in iterate_schema(
        process_input, input_schema, include_groups=True
    ):
        if "group" in field_schema and field_schema["name"] not in fields:
            dict_dot(process_input, path, {})
        if "default" in field_schema and field_schema["name"] not in fields:
            dict_dot(process_input, path, field_schema["default"])



# Generated by Django 2.2.4 on 2019-08-22 17:30

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0039_entity_collection_cascade"),
    ]

    operations = [
        migrations.RemoveField(model_name="entity", name="descriptor_completed",),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2018-08-06 02:16
from __future__ import unicode_literals

from django.db import migrations


def position_to_label(apps, schema_editor):
    RelationPartition = apps.get_model("flow", "RelationPartition")
    for partition in RelationPartition.objects.all():
        if partition.position:
            partition.label = partition.position
        partition.save()


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0015_refactor_relations_1"),
    ]

    operations = [
        migrations.RunPython(position_to_label),
    ]



"""Resolwe descriptor schema model."""
from django.contrib.postgres.fields import JSONField
from django.db import models

from .base import BaseModel


class DescriptorSchema(BaseModel):
    """Postgres model for storing descriptors."""

    class Meta(BaseModel.Meta):
        """DescriptorSchema Meta options."""

        permissions = (
            ("view_descriptorschema", "Can view descriptor schema"),
            ("edit_descriptorschema", "Can edit descriptor schema"),
            ("share_descriptorschema", "Can share descriptor schema"),
            ("owner_descriptorschema", "Is owner of the description schema"),
        )

    #: detailed description
    description = models.TextField(blank=True)

    #: user descriptor schema represented as a JSON object
    schema = JSONField(default=list)



"""Logging configuration for executors."""
import asyncio
import json
import logging
import os
import socket
import sys
from logging.config import dictConfig

from .global_settings import DATA, DATA_LOCATION
from .manager_commands import send_manager_command
from .protocol import ExecutorProtocol


class JSONFormatter(logging.Formatter):
    """Logger formatter for dumping records to JSON."""

    def format(self, record):
        """Dump the record to JSON."""
        data = record.__dict__.copy()

        data["data_id"] = DATA["id"]
        data["data_location_id"] = DATA_LOCATION["id"]
        data["hostname"] = socket.gethostname()

        # Get relative path, so listener can reconstruct the path to the actual code.
        data["pathname"] = os.path.relpath(data["pathname"], os.path.dirname(__file__))

        # Exception and Traceback cannot be serialized.
        data["exc_info"] = None

        # Ensure logging message is instantiated to a string.
        data["msg"] = str(data["msg"])

        return json.dumps(data)


class RedisHandler(logging.Handler):
    """Publish messages to Redis channel."""

    def __init__(self, emit_list, **kwargs):
        """Construct a handler instance.

        :param emit_list: The list to add emit futures into, so they can
            be waited on in the executor main function.
        """
        self.emit_list = emit_list
        super().__init__(**kwargs)

    def emit(self, record):
        """Send log message to the listener."""
        future = asyncio.ensure_future(
            send_manager_command(
                ExecutorProtocol.LOG,
                extra_fields={ExecutorProtocol.LOG_MESSAGE: self.format(record),},
                expect_reply=False,
            )
        )
        self.emit_list.append(future)


def configure_logging(emit_list):
    """Configure logging to send log records to the master."""
    if "sphinx" in sys.modules:
        module_base = "resolwe.flow.executors"
    else:
        module_base = "executors"
    logging_config = dict(
        version=1,
        formatters={"json_formatter": {"()": JSONFormatter},},
        handlers={
            "redis": {
                "class": module_base + ".logger.RedisHandler",
                "formatter": "json_formatter",
                "level": logging.INFO,
                "emit_list": emit_list,
            },
            "console": {"class": "logging.StreamHandler", "level": logging.WARNING},
        },
        root={"handlers": ["redis", "console"], "level": logging.DEBUG,},
        loggers={
            # Don't use redis logger to prevent circular dependency.
            module_base
            + ".manager_comm": {
                "level": "INFO",
                "handlers": ["console"],
                "propagate": False,
            },
        },
    )

    dictConfig(logging_config)



""".. Ignore pydocstyle D400.

===================================
Resolwe Test Helpers and Decorators
===================================

"""
import functools
import os
import shlex
import shutil
import subprocess
import unittest
from sys import platform

import wrapt

from django.conf import settings
from django.test import override_settings, tag

__all__ = (
    "check_installed",
    "check_docker",
    "with_custom_executor",
    "with_docker_executor",
    "with_null_executor",
    "with_resolwe_host",
    "is_testing",
)

TAG_PROCESS = "resolwe.process"


def check_installed(command):
    """Check if the given command is installed.

    :param str command: name of the command

    :return: (indicator of the availability of the command, message saying
              command is not available)
    :rtype: tuple(bool, str)

    """
    if shutil.which(command):
        return True, ""
    else:
        return False, "Command '{}' is not found.".format(command)


def check_docker():
    """Check if Docker is installed and working.

    :return: (indicator of the availability of Docker, reason for
              unavailability)
    :rtype: tuple(bool, str)

    """
    command = getattr(settings, "FLOW_DOCKER_COMMAND", "docker")
    info_command = "{} info".format(command)
    available, reason = True, ""
    # TODO: Use subprocess.DEVNULL after dropping support for Python 2
    with open(os.devnull, "wb") as DEVNULL:
        try:
            subprocess.check_call(
                shlex.split(info_command), stdout=DEVNULL, stderr=subprocess.STDOUT
            )
        except OSError:
            available, reason = False, "Docker command '{}' not found".format(command)
        except subprocess.CalledProcessError:
            available, reason = (
                False,
                "Docker command '{}' returned non-zero "
                "exit status".format(info_command),
            )
    return available, reason


def with_custom_executor(wrapped=None, **custom_executor_settings):
    """Decorate unit test to run processes with a custom executor.

    :param dict custom_executor_settings: custom ``FLOW_EXECUTOR``
        settings with which you wish to override the current settings

    """
    if wrapped is None:
        return functools.partial(with_custom_executor, **custom_executor_settings)

    @wrapt.decorator
    def wrapper(wrapped_method, instance, args, kwargs):
        from resolwe.flow.managers import manager  # To prevent circular imports.

        executor_settings = settings.FLOW_EXECUTOR.copy()
        executor_settings.update(custom_executor_settings)

        try:
            with override_settings(FLOW_EXECUTOR=executor_settings):
                with manager.override_settings(FLOW_EXECUTOR=executor_settings):
                    # Re-run engine discovery as the settings have changed.
                    manager.discover_engines()

                    # Re-run the post_register_hook
                    manager.get_executor().post_register_hook(verbosity=0)

                    # Run the actual unit test method.
                    return wrapped_method(*args, **kwargs)
        finally:
            # Re-run engine discovery as the settings have changed.
            manager.discover_engines()

    return wrapper(wrapped)


def with_docker_executor(wrapped=None):
    """Decorate unit test to run processes with the Docker executor."""
    if wrapped is None:
        return functools.partial(with_docker_executor)

    @wrapt.decorator
    def wrapper(wrapped_method, instance, args, kwargs):
        return unittest.skipUnless(*check_docker())(
            with_custom_executor(NAME="resolwe.flow.executors.docker",)(wrapped_method)
        )(*args, **kwargs)

    return wrapper(wrapped)


@wrapt.decorator
def with_null_executor(wrapped_method, instance, args, kwargs):
    """Decorate unit test to run processes with the Null executor."""
    return with_custom_executor(NAME="resolwe.flow.executors.null")(wrapped_method)(
        *args, **kwargs
    )


@wrapt.decorator
def with_resolwe_host(wrapped_method, instance, args, kwargs):
    """Decorate unit test to give it access to a live Resolwe host.

    Set ``RESOLWE_HOST_URL`` setting to the address where the testing
    live Resolwe host listens to.

    .. note::

        This decorator must be used with a (sub)class of
        :class:`~django.test.LiveServerTestCase` which starts a live
        Django server in the background.

    """
    from resolwe.flow.managers import manager  # To prevent circular imports.

    if not hasattr(instance, "server_thread"):
        raise AttributeError(
            "with_resolwe_host decorator must be used with a "
            "(sub)class of LiveServerTestCase that has the "
            "'server_thread' attribute"
        )
    host = instance.server_thread.host
    if platform not in ["linux", "linux2"]:
        host = "host.docker.internal"
    resolwe_host_url = "http://{host}:{port}".format(
        host=host, port=instance.server_thread.port
    )
    with override_settings(RESOLWE_HOST_URL=resolwe_host_url):
        with manager.override_settings(RESOLWE_HOST_URL=resolwe_host_url):
            # Run the actual unit test method.
            return with_custom_executor(NETWORK="host")(wrapped_method)(*args, **kwargs)


def generate_process_tag(slug):
    """Generate test tag for a given process."""
    return "{}.{}".format(TAG_PROCESS, slug)


def tag_process(*slugs):
    """Decorate unit test to tag it for a specific process."""
    slugs = [generate_process_tag(slug) for slug in slugs]
    slugs.append(TAG_PROCESS)  # Also tag with a general process tag.
    return tag(*slugs)


def has_process_tag(test, slug):
    """Check if a given test method/class is tagged for a specific process."""
    tags = getattr(test, "tags", set())
    return generate_process_tag(slug) in tags


def get_processes_from_tags(test):
    """Extract process slugs from tags."""
    tags = getattr(test, "tags", set())
    slugs = set()

    for tag_name in tags:
        if not tag_name.startswith("{}.".format(TAG_PROCESS)):
            continue

        slugs.add(tag_name[len(TAG_PROCESS) + 1 :])

    return slugs


def is_testing():
    """Return current testing status.

    This assumes that the Resolwe test runner is being used.
    """
    from resolwe.test_helpers.test_runner import is_testing

    return is_testing()



# pylint: disable=missing-docstring
import copy

from django.conf import settings

from resolwe.flow.managers import manager
from resolwe.flow.models import Data, Process
from resolwe.test import TransactionTestCase, with_docker_executor


class EnvVarsTest(TransactionTestCase):
    @with_docker_executor
    def test_envvars(self):
        flow_executor = copy.copy(getattr(settings, "FLOW_EXECUTOR", {}))
        flow_executor["SET_ENV"] = {
            "SET_ENV_TEST": "test_var",
        }

        with manager.override_settings(
            FLOW_EXECUTOR=flow_executor, RESOLWE_HOST_URL="some.special.host"
        ):
            process = Process.objects.create(
                name="Test environment variables",
                requirements={"expression-engine": "jinja"},
                contributor=self.contributor,
                type="test:data:envvars:",
                input_schema=[],
                output_schema=[
                    {"name": "resolweapihost", "type": "basic:string:"},
                    {"name": "setenvtest", "type": "basic:string:"},
                    {"name": "tmpdir", "type": "basic:string:"},
                ],
                run={
                    "language": "bash",
                    "program": """
re-save resolweapihost $RESOLWE_HOST_URL
re-save setenvtest $SET_ENV_TEST
re-save tmpdir $TMPDIR
""",
                },
            )

            data = Data.objects.create(
                name="Data object",
                contributor=self.contributor,
                process=process,
                input={},
            )

            # update output
            data = Data.objects.get(pk=data.pk)

            self.assertEqual(data.output["resolweapihost"], "some.special.host")
            self.assertEqual(data.output["setenvtest"], "test_var")
            self.assertEqual(data.output["tmpdir"], "/data/.tmp")



# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2018-08-06 02:13
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0014_track_resources"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="PositionInRelation", new_name="RelationPartition",
        ),
        migrations.RenameField(
            model_name="relation", old_name="label", new_name="category",
        ),
        migrations.AlterField(
            model_name="relation",
            name="category",
            field=models.CharField(default="", max_length=100),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="relation",
            name="unit",
            field=models.CharField(
                blank=True,
                choices=[
                    ("s", "Second"),
                    ("min", "Minute"),
                    ("hr", "Hour"),
                    ("d", "Day"),
                    ("wk", "Week"),
                ],
                max_length=3,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="relation",
            name="collection",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="flow.Collection"
            ),
        ),
        migrations.AlterUniqueTogether(
            name="relation", unique_together=set([("collection", "category")]),
        ),
        migrations.AddField(
            model_name="relationpartition",
            name="label",
            field=models.CharField(blank=True, db_index=True, max_length=30, null=True),
        ),
        migrations.AlterUniqueTogether(
            name="relationpartition", unique_together=set([("entity", "relation")]),
        ),
    ]



""".. Ignore pydocstyle D400.

==========
Flow Views
==========

.. autoclass:: resolwe.flow.views.collection.CollectionViewSet
    :members:

.. autoclass:: resolwe.flow.views.data.DataViewSet
    :members:

.. autoclass:: resolwe.flow.views.descriptor.DescriptorSchemaViewSet
    :members:

.. autoclass:: resolwe.flow.views.entity.EntityViewSet
    :members:

.. autoclass:: resolwe.flow.views.process.ProcessViewSet
    :members:

.. autoclass:: resolwe.flow.views.relation.RelationViewSet
    :members:

.. autoclass:: resolwe.flow.views.storage.StorageViewSet
    :members:

"""
from .collection import CollectionViewSet
from .data import DataViewSet
from .descriptor import DescriptorSchemaViewSet
from .entity import EntityViewSet
from .process import ProcessViewSet
from .relation import RelationViewSet
from .storage import StorageViewSet

__all__ = (
    "CollectionViewSet",
    "DataViewSet",
    "DescriptorSchemaViewSet",
    "EntityViewSet",
    "ProcessViewSet",
    "RelationViewSet",
    "StorageViewSet",
)



# pylint: disable=missing-docstring
import datetime

from django.contrib.auth import get_user_model
from django.utils.timezone import get_current_timezone

from guardian.shortcuts import assign_perm
from rest_framework import status
from rest_framework.test import APIRequestFactory, force_authenticate

from resolwe.flow.filters import CollectionFilter, DataFilter, EntityFilter
from resolwe.flow.models import Collection, Data, DescriptorSchema, Entity, Process
from resolwe.flow.views import (
    CollectionViewSet,
    DataViewSet,
    DescriptorSchemaViewSet,
    ProcessViewSet,
)
from resolwe.test import TestCase

factory = APIRequestFactory()


class BaseViewSetFiltersTest(TestCase):
    def setUp(self):
        super().setUp()

        self.viewset = self.viewset_class.as_view(actions={"get": "list",})

    def _check_filter(
        self, query_args, expected, expected_status_code=status.HTTP_200_OK
    ):
        """Check that query_args filter to expected queryset."""
        request = factory.get("/", query_args, format="json")
        force_authenticate(request, self.admin)
        response = self.viewset(request)

        if status.is_success(response.status_code):
            self.assertEqual(len(response.data), len(expected))
            self.assertCountEqual(
                [item.pk for item in expected], [item["id"] for item in response.data],
            )
        else:
            self.assertEqual(response.status_code, expected_status_code)
            response.render()
            return response


class CollectionViewSetFiltersTest(BaseViewSetFiltersTest):
    def setUp(self):
        self.viewset_class = CollectionViewSet
        super().setUp()

        self.collections = []
        for index in range(2):
            collection = Collection.objects.create(
                name="Test collection {}".format(index),
                slug="test-collection-{}".format(index),
                contributor=self.contributor,
                description="My description!" if index == 0 else "Other description",
            )

            assign_perm("owner_collection", self.admin, collection)

            self.collections.append(collection)

        collection = Collection.objects.create(
            name="User test collection",
            slug="user-collection",
            contributor=self.user,
            description="User's description",
        )

        assign_perm("owner_collection", self.user, collection)

        self.collections.append(collection)

    def test_filter_name(self):
        self._check_filter({"name": "Test collection 1"}, [self.collections[1]])
        self._check_filter({"name": "Test collection"}, [])

    def test_filter_name_contains(self):
        self._check_filter({"name_contains": "test"}, self.collections)
        self._check_filter({"name_contains": "xxx"}, [])

    def test_filter_contributor_name(self):
        self._check_filter(
            {"contributor_name": "Joe"}, [self.collections[0], self.collections[1]]
        )
        self._check_filter({"contributor_name": "John"}, [self.collections[2]])
        self._check_filter({"contributor_name": "xxx"}, [])

    def test_filter_owners_name(self):
        self._check_filter(
            {"owners_name": "James"}, [self.collections[0], self.collections[1]]
        )
        self._check_filter({"owners_name": "John"}, [self.collections[2]])
        self._check_filter({"owners_name": "Williams"}, [self.collections[2]])
        self._check_filter({"owners_name": "xxx"}, [])

    def test_filter_text(self):
        # By slug.
        self._check_filter({"text": "test-collection-1"}, [self.collections[1]])
        self._check_filter({"text": "test-colle"}, self.collections)

        # By name.
        self._check_filter({"text": "Test collection 1"}, [self.collections[1]])
        self._check_filter({"text": "Test"}, self.collections)
        self._check_filter({"text": "test"}, self.collections)
        self._check_filter({"text": "tes"}, self.collections)
        self._check_filter({"text": "user"}, [self.collections[2]])
        self._check_filter({"text": "est"}, [])

        # By contributor.
        self._check_filter({"text": "joe"}, [self.collections[0], self.collections[1]])
        self._check_filter({"text": "oe"}, [])
        self._check_filter(
            {"text": "Miller"}, [self.collections[0], self.collections[1]]
        )
        self._check_filter({"text": "mill"}, [self.collections[0], self.collections[1]])
        self._check_filter({"text": "john"}, [self.collections[2]])

        # By owner.
        self._check_filter(
            {"text": "james"}, [self.collections[0], self.collections[1]]
        )
        self._check_filter({"text": "mes"}, [])
        self._check_filter(
            {"text": "Smith"}, [self.collections[0], self.collections[1]]
        )
        self._check_filter({"text": "smi"}, [self.collections[0], self.collections[1]])
        self._check_filter({"text": "john"}, [self.collections[2]])

        # By description.
        self._check_filter({"text": "description"}, self.collections)
        self._check_filter({"text": "my"}, [self.collections[0]])
        self._check_filter({"text": "my description"}, [self.collections[0]])
        self._check_filter({"text": "user"}, [self.collections[2]])
        self._check_filter({"text": "ription"}, [])


class DataViewSetFiltersTest(BaseViewSetFiltersTest):
    def setUp(self):
        self.viewset_class = DataViewSet
        super().setUp()

        tzone = get_current_timezone()

        self.collection = Collection.objects.create(contributor=self.contributor)

        self.proc1 = Process.objects.create(
            type="data:test:process1:",
            name="First process",
            slug="test-process-1",
            version="1.0.0",
            contributor=self.contributor,
            entity_type="test-schema",
            entity_descriptor_schema="test-schema",
            input_schema=[
                {"name": "input_data", "type": "data:test:", "required": False}
            ],
        )

        self.proc2 = Process.objects.create(
            type="data:test:process2:",
            name="Second process",
            slug="test-process-2",
            version="1.0.0",
            contributor=self.contributor,
            entity_type="test-schema",
            entity_descriptor_schema="test-schema",
            input_schema=[
                {"name": "input_data", "type": "data:test:", "required": False}
            ],
        )

        self.descriptor_schema = DescriptorSchema.objects.create(
            slug="test-schema", version="1.0.0", contributor=self.contributor,
        )

        self.data = []
        for index in range(10):
            data = Data.objects.create(
                name="Data {}".format(index),
                slug="dataslug-{}".format(index),
                contributor=self.contributor if index < 9 else self.user,
                process=self.proc1 if index < 5 else self.proc2,
                status=Data.STATUS_DONE if index > 0 else Data.STATUS_RESOLVING,
                started=datetime.datetime(2016, 7, 31, index, 0, tzinfo=tzone),
                finished=datetime.datetime(2016, 7, 31, index, 30, tzinfo=tzone),
                tags=["foo", "index{}".format(index)],
            )

            data.created = datetime.datetime(2016, 7, 30, index, 59, tzinfo=tzone)
            data.save()

            assign_perm("owner_data", self.admin if index < 9 else self.user, data)

            if index == 0:
                data.collection = self.collection
                data.save()

            self.data.append(data)

    def test_filter_id(self):
        self._check_filter({"id": str(self.data[0].pk)}, [self.data[0]])
        self._check_filter({"id": str(self.data[2].pk)}, [self.data[2]])
        self._check_filter(
            {"id__in": "{},{}".format(self.data[0].pk, self.data[2].pk)},
            [self.data[0], self.data[2]],
        )

    def test_filter_slug(self):
        self._check_filter({"slug": "dataslug-1"}, [self.data[1]])
        self._check_filter({"slug": "dataslug-5"}, [self.data[5]])
        self._check_filter({"slug__in": "dataslug-1"}, [self.data[1]])
        self._check_filter(
            {"slug__in": "dataslug-1,dataslug-5"}, [self.data[1], self.data[5]]
        )

    def test_filter_name(self):
        self._check_filter({"name": "Data 1"}, [self.data[1]])
        self._check_filter({"name": "data 1"}, [])
        self._check_filter({"name": "Data 2"}, [self.data[2]])
        self._check_filter({"name": "Data"}, [])
        self._check_filter({"name": "1"}, [])

    def test_filter_name_contains(self):
        self._check_filter({"name_contains": "0"}, self.data[:1])
        self._check_filter({"name_contains": "data"}, self.data)
        self._check_filter({"name_contains": "xxx"}, [])

    def test_filter_contributor(self):
        self._check_filter({"contributor": str(self.contributor.pk)}, self.data[:9])
        self._check_filter({"contributor": "0"}, [])
        self._check_filter(
            {"contributor__in": "0,{}".format(self.contributor.pk)}, self.data[:9]
        )

    def test_filter_contributor_name(self):
        self._check_filter({"contributor_name": "Joe"}, self.data[:9])
        self._check_filter({"contributor_name": "John"}, self.data[9:])
        self._check_filter({"contributor_name": "xxx"}, [])

    def test_filter_owners(self):
        self._check_filter({"owners": str(self.admin.pk)}, self.data[:9])
        self._check_filter({"owners": "0"}, [])
        self._check_filter({"owners__in": "0,{}".format(self.admin.pk)}, self.data[:9])

    def test_filter_owners_name(self):
        self._check_filter({"owners_name": "Joe"}, self.data[:9])
        self._check_filter({"owners_name": "John"}, self.data[9:])
        self._check_filter({"owners_name": "xxx"}, [])

    def test_filter_created(self):
        self._check_filter(
            {"created": self.data[0].created.isoformat()}, [self.data[0]]
        )
        self._check_filter({"created__gt": "2016"}, self.data)
        self._check_filter({"created__gt": "2016-07-30T05:59:00"}, self.data[6:])
        self._check_filter({"created__gte": "2016-07-30T05:59:00"}, self.data[5:])
        self._check_filter({"created__lt": "2016"}, [])
        self._check_filter({"created__lt": "2016-07-30T05:59:00"}, self.data[:5])
        self._check_filter({"created__lte": "2016-07-30T05:59:00"}, self.data[:6])

    def test_filter_modified(self):
        now = self.data[0].modified
        five_offset = self.data[5].modified
        self._check_filter({"modified": now.isoformat()}, [self.data[0]])
        self._check_filter({"modified__gt": str(now.year)}, self.data)
        self._check_filter({"modified__gt": five_offset.isoformat()}, self.data[6:])
        self._check_filter({"modified__gte": five_offset.isoformat()}, self.data[5:])
        self._check_filter({"modified__lt": str(now.year)}, [])
        self._check_filter({"modified__lt": five_offset.isoformat()}, self.data[:5])
        self._check_filter({"modified__lte": five_offset.isoformat()}, self.data[:6])

    def test_filter_started(self):
        self._check_filter(
            {"started": self.data[0].started.isoformat()}, [self.data[0]]
        )
        self._check_filter({"started__gt": "2016"}, self.data)
        self._check_filter({"started__gt": "2016-07-31T05:00:00"}, self.data[6:])
        self._check_filter({"started__gte": "2016-07-31T05:00:00"}, self.data[5:])
        self._check_filter({"started__lt": "2016"}, [])
        self._check_filter({"started__lt": "2016-07-31T05:00:00"}, self.data[:5])
        self._check_filter({"started__lte": "2016-07-31T05:00:00"}, self.data[:6])

    def test_filter_finished(self):
        self._check_filter(
            {"finished": self.data[0].finished.isoformat()}, [self.data[0]]
        )
        self._check_filter({"finished__gt": "2016"}, self.data)
        self._check_filter({"finished__gt": "2016-07-31T05:30:00"}, self.data[6:])
        self._check_filter({"finished__gte": "2016-07-31T05:30:00"}, self.data[5:])
        self._check_filter({"finished__lt": "2016"}, [])
        self._check_filter({"finished__lt": "2016-07-31T05:30:00"}, self.data[:5])
        self._check_filter({"finished__lte": "2016-07-31T05:30:00"}, self.data[:6])

    def test_filter_collection(self):
        self._check_filter({"collection": str(self.collection.pk)}, [self.data[0]])

    def test_filter_type(self):
        self._check_filter({"type": "data"}, self.data)
        self._check_filter({"type": "data:test"}, self.data)
        self._check_filter({"type": "data:test:process1"}, self.data[:5])
        self._check_filter({"type": "data:test:process2"}, self.data[5:])
        self._check_filter({"type": "data:"}, [])
        self._check_filter({"type": "data:test:"}, [])
        self._check_filter({"type": "data:test:process1:"}, self.data[:5])
        self._check_filter({"type": "data:test:process2:"}, self.data[5:])
        self._check_filter({"type__exact": "data:test"}, [])
        self._check_filter({"type__exact": "data:test:process1:"}, self.data[:5])
        self._check_filter({"type__exact": "data:test:process2:"}, self.data[5:])

    def test_filter_status(self):
        self._check_filter({"status": "OK"}, self.data[1:])
        self._check_filter({"status": "RE"}, [self.data[0]])
        self._check_filter({"status__in": "OK,RE"}, self.data)

    def test_filter_process(self):
        self._check_filter({"process": str(self.proc1.pk)}, self.data[:5])
        self._check_filter({"process": str(self.proc2.pk)}, self.data[5:])

    def test_filter_process_name(self):
        self._check_filter({"process_name": "first"}, self.data[:5])
        self._check_filter({"process_name": "fir"}, self.data[:5])
        self._check_filter({"process_name": "rst"}, [])
        self._check_filter({"process_name": "second"}, self.data[5:])
        self._check_filter({"process_name": "sec"}, self.data[5:])

    def test_filter_process_slug(self):
        self._check_filter({"process_slug": "test-process-1"}, self.data[:5])
        self._check_filter({"process_slug": "test-process-2"}, self.data[5:])
        self._check_filter({"process_slug": "test-process"}, [])

    def test_filter_tags(self):
        self._check_filter({"tags": "foo"}, self.data)
        self._check_filter({"tags": "foo,index1"}, [self.data[1]])
        self._check_filter({"tags": "foo,index5"}, [self.data[5]])
        self._check_filter({"tags": "foo,index10"}, [])
        self._check_filter({"tags": "bar"}, [])

    def test_filter_text(self):
        # By slug.
        self._check_filter({"text": "dataslug-1"}, [self.data[1]])
        self._check_filter({"text": "datasl"}, self.data)

        # By name.
        self._check_filter({"text": "Data 1"}, [self.data[1]])
        self._check_filter({"text": "data 2"}, [self.data[2]])
        self._check_filter({"text": "Data"}, self.data)
        self._check_filter({"text": "data"}, self.data)
        self._check_filter({"text": "dat"}, self.data)
        self._check_filter({"text": "ata"}, [])

        # By contributor.
        self._check_filter({"text": "joe"}, self.data[:9])
        self._check_filter({"text": "oe"}, [])
        self._check_filter({"text": "Miller"}, self.data[:9])
        self._check_filter({"text": "mill"}, self.data[:9])
        self._check_filter({"text": "john"}, self.data[9:])

        # By owner.
        self._check_filter({"text": "james"}, self.data[:9])
        self._check_filter({"text": "mes"}, [])
        self._check_filter({"text": "Smith"}, self.data[:9])
        self._check_filter({"text": "smi"}, self.data[:9])

        # By process name.
        self._check_filter({"text": "first"}, self.data[:5])
        self._check_filter({"text": "fir"}, self.data[:5])
        self._check_filter({"text": "rst"}, [])
        self._check_filter({"text": "process"}, self.data)

    def test_nonexisting_parameter(self):
        response = self._check_filter(
            {"foo": "bar"}, [self.data[0]], expected_status_code=400
        )

        self.assertRegex(
            str(response.data["detail"]),
            r"Unsupported parameter\(s\): foo. Please use a combination of: .*",
        )


class DescriptorSchemaViewSetFiltersTest(BaseViewSetFiltersTest):
    def setUp(self):
        self.viewset_class = DescriptorSchemaViewSet
        super().setUp()

        tzone = get_current_timezone()

        self.ds1 = DescriptorSchema.objects.create(
            contributor=self.user, slug="slug1", name="ds1"
        )
        self.ds1.created = datetime.datetime(2015, 7, 28, 11, 57, tzinfo=tzone)
        self.ds1.save()

        self.ds2 = DescriptorSchema.objects.create(
            contributor=self.user, slug="slug2", name="ds2"
        )
        self.ds2.created = datetime.datetime(2016, 8, 29, 12, 58, 0, tzinfo=tzone)
        self.ds2.save()

        self.ds3 = DescriptorSchema.objects.create(
            contributor=self.admin, slug="slug3", name="ds3"
        )
        self.ds3.created = datetime.datetime(2017, 9, 30, 13, 59, 0, tzinfo=tzone)
        self.ds3.save()

    def test_id(self):
        self._check_filter({"id": self.ds1.pk}, [self.ds1])
        self._check_filter({"id": self.ds2.pk}, [self.ds2])
        self._check_filter({"id__in": self.ds2.pk}, [self.ds2])
        self._check_filter(
            {"id__in": "{},{}".format(self.ds1.pk, self.ds2.pk)}, [self.ds1, self.ds2]
        )
        self._check_filter({"id__gt": self.ds1.pk}, [self.ds2, self.ds3])

    def test_slug(self):
        self._check_filter({"slug": "slug1"}, [self.ds1])
        self._check_filter({"slug": "slug2"}, [self.ds2])
        self._check_filter({"slug__icontains": "LUG"}, [self.ds1, self.ds2, self.ds3])
        self._check_filter({"slug__in": "slug"}, [])
        self._check_filter({"slug__in": "slug2,slug"}, [self.ds2])
        self._check_filter({"slug__startswith": "slug"}, [self.ds1, self.ds2, self.ds3])
        self._check_filter({"slug__startswith": "slug2"}, [self.ds2])
        self._check_filter({"slug__endswith": "g1"}, [self.ds1])

    def test_name(self):
        self._check_filter({"name": "ds1"}, [self.ds1])
        self._check_filter({"name": "ds2"}, [self.ds2])
        self._check_filter({"name__startswith": "ds"}, [self.ds1, self.ds2, self.ds3])
        self._check_filter({"name__startswith": "ds2"}, [self.ds2])
        self._check_filter({"name__startswith": "dsx"}, [])
        self._check_filter({"name__icontains": "2"}, [self.ds2])
        self._check_filter({"name__icontains": "DS"}, [self.ds1, self.ds2, self.ds3])

    def test_contributor(self):
        self._check_filter({"contributor": self.user.pk}, [self.ds1, self.ds2])
        self._check_filter({"contributor": self.admin.pk}, [self.ds3])
        self._check_filter({"contributor__in": self.admin.pk}, [self.ds3])
        self._check_filter(
            {"contributor__in": "{},{}".format(self.user.pk, self.admin.pk)},
            [self.ds1, self.ds2, self.ds3],
        )

    def test_created(self):
        self._check_filter({"created": "2015-07-28T11:57:00"}, [self.ds1])
        self._check_filter({"created": "2016-08-29T12:58:00.000000"}, [self.ds2])
        self._check_filter({"created": "2017-09-30T13:59:00.000000Z"}, [self.ds3])
        self._check_filter({"created": "2017-09-30T13:59:00.000000+0000"}, [self.ds3])
        self._check_filter({"created": "2017-09-30T13:59:00.000000+0000"}, [self.ds3])

        self._check_filter({"created__date": "2016-08-29"}, [self.ds2])
        self._check_filter({"created__time": "12:58:00"}, [self.ds2])

        self._check_filter({"created__year": "2016"}, [self.ds2])
        self._check_filter({"created__year__gt": "2015"}, [self.ds2, self.ds3])
        self._check_filter({"created__month__gte": "8"}, [self.ds2, self.ds3])
        self._check_filter({"created__day__lt": "29"}, [self.ds1])
        self._check_filter({"created__hour__lte": "12"}, [self.ds1, self.ds2])
        self._check_filter({"created__minute__gt": "57"}, [self.ds2, self.ds3])

    def test_modified(self):
        now = self.ds1.modified
        self._check_filter(
            {"modified": now.strftime("%Y-%m-%dT%H:%M:%S.%f%z")}, [self.ds1]
        )
        self._check_filter({"modified__year": now.year}, [self.ds1, self.ds2, self.ds3])

    def test_nonexisting_parameter(self):
        response = self._check_filter(
            {"foo": "bar"}, [self.ds1], expected_status_code=400
        )

        self.assertRegex(
            str(response.data["__all__"][0]),
            r"Unsupported parameter\(s\): foo. Please use a combination of: .*",
        )


class ProcessViewSetFiltersTest(BaseViewSetFiltersTest):
    def setUp(self):
        self.viewset_class = ProcessViewSet
        super().setUp()

        self.proc_1 = Process.objects.create(
            contributor=self.contributor,
            type="data:alignment:bam:",
            category="analyses:alignment:",
            scheduling_class="BA",
        )

        self.proc_2 = Process.objects.create(
            contributor=self.contributor,
            type="data:expression:",
            category="analyses:",
            scheduling_class="IN",
        )

    def test_category(self):
        self._check_filter({"category": "analyses:"}, [self.proc_1, self.proc_2])
        self._check_filter({"category": "analyses:alignment"}, [self.proc_1])

    def test_type(self):
        self._check_filter({"type": "data:"}, [self.proc_1, self.proc_2])
        self._check_filter({"type": "data:alignment:bam"}, [self.proc_1])
        self._check_filter({"type": "data:expression"}, [self.proc_2])

    def test_scheduling_class(self):
        self._check_filter({"scheduling_class": "BA"}, [self.proc_1])
        self._check_filter({"scheduling_class": "IN"}, [self.proc_2])

    def test_is_active(self):
        self._check_filter({"is_active": "true"}, [self.proc_1, self.proc_2])
        self._check_filter({"is_active": "false"}, [])

        self.proc_1.is_active = False
        self.proc_1.save()

        self._check_filter({"is_active": "true"}, [self.proc_2])
        self._check_filter({"is_active": "false"}, [self.proc_1])


class DataFilterTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user_1 = get_user_model().objects.create(username="first_user")
        cls.user_2 = get_user_model().objects.create(username="second_user")

        tzone = get_current_timezone()

        cls.proc_1 = Process.objects.create(
            name="Test process 1",
            slug="test-process-1",
            contributor=cls.user_1,
            type="data:test:",
        )

        cls.proc_2 = Process.objects.create(
            name="Test process 2",
            slug="test-process-2",
            contributor=cls.user_1,
            type="data:test:second:",
        )

        cls.proc_3 = Process.objects.create(
            name="Test process 3",
            slug="test-process-3",
            contributor=cls.user_1,
            type="data:third:",
        )

        cls.data_1 = Data.objects.create(
            name="Test data 1",
            slug="test-data-1",
            contributor=cls.user_1,
            process=cls.proc_1,
            status=Data.STATUS_DONE,
            started=datetime.datetime(2016, 7, 30, 14, 0, tzinfo=tzone),
            finished=datetime.datetime(2016, 7, 30, 14, 30, tzinfo=tzone),
            size=42,
        )
        cls.data_1.created = datetime.datetime(2016, 7, 30, 13, 59, tzinfo=tzone)
        cls.data_1.save()

        cls.data_2 = Data.objects.create(
            name="Test data 2",
            slug="test-data-2",
            contributor=cls.user_1,
            process=cls.proc_2,
            status=Data.STATUS_ERROR,
            started=datetime.datetime(2016, 8, 30, 15, 0, tzinfo=tzone),
            finished=datetime.datetime(2016, 8, 30, 15, 30, tzinfo=tzone),
            tags=["foo", "bar", "moo"],
            size=42,
        )
        cls.data_2.created = datetime.datetime(2016, 8, 30, 14, 59, tzinfo=tzone)
        cls.data_2.save()

        cls.data_3 = Data.objects.create(
            name="Another data object",
            slug="another-data-object",
            contributor=cls.user_2,
            process=cls.proc_3,
            status=Data.STATUS_DONE,
            started=datetime.datetime(2013, 1, 15, 8, 0, tzinfo=tzone),
            finished=datetime.datetime(2013, 1, 15, 10, 0, tzinfo=tzone),
            tags=["bar"],
        )
        cls.data_3.created = datetime.datetime(2013, 1, 15, 7, 59, tzinfo=tzone)
        cls.data_3.save()

        cls.collection = Collection.objects.create(
            name="Test collection", slug="test-collection", contributor=cls.user_1
        )
        cls.collection.data.add(cls.data_1)

    def _apply_filter(self, filters, expected):
        filtered = DataFilter(filters, queryset=Data.objects.all())
        self.assertCountEqual(filtered.qs, expected)

    def test_filter_id(self):
        self._apply_filter({"id": self.data_1.pk}, [self.data_1])
        self._apply_filter(
            {"id__in": "{},{}".format(self.data_1.pk, self.data_2.pk)},
            [self.data_1, self.data_2],
        )

    def test_filter_slug(self):
        self._apply_filter({"slug": "test-data-2"}, [self.data_2])
        self._apply_filter({"slug__startswith": "test"}, [self.data_1, self.data_2])

    def test_filter_name(self):
        self._apply_filter({"name": "Test data 2"}, [self.data_2])
        self._apply_filter({"name__startswith": "Test"}, [self.data_1, self.data_2])

    def test_filter_contributor(self):
        self._apply_filter({"contributor": self.user_1.pk}, [self.data_1, self.data_2])

    def test_filter_created(self):
        self._apply_filter({"created__year": "2016"}, [self.data_1, self.data_2])
        self._apply_filter({"created__year__gt": "2015"}, [self.data_1, self.data_2])
        self._apply_filter(
            {"created__year__gt": "2015", "created__month__gte": "8"}, [self.data_2]
        )

    def test_filter_modified(self):
        year = datetime.date.today().strftime("%Y")
        self._apply_filter(
            {"modified__year": year}, [self.data_1, self.data_2, self.data_3]
        )
        self._apply_filter({"modified__year__gt": year}, [])

    def test_filter_collection(self):
        self._apply_filter({"collection": self.collection.pk}, [self.data_1])
        self._apply_filter({"collection__slug": "test-collection"}, [self.data_1])

    def test_filter_type(self):
        self._apply_filter({"type": "data:test:"}, [self.data_1, self.data_2])

    def test_filter_status(self):
        self._apply_filter({"status": "ok"}, [self.data_1, self.data_3])
        self._apply_filter({"status__in": "ER"}, [self.data_2])
        self._apply_filter(
            {"status__in": "ER,OK"}, [self.data_1, self.data_2, self.data_3]
        )

    def test_filter_finished(self):
        self._apply_filter({"finished__year": "2016"}, [self.data_1, self.data_2])
        self._apply_filter({"finished__year__gt": "2015"}, [self.data_1, self.data_2])
        self._apply_filter(
            {"finished__year__gt": "2015", "finished__month__gte": "8"}, [self.data_2]
        )

    def test_filter_started(self):
        self._apply_filter({"started__year": "2016"}, [self.data_1, self.data_2])
        self._apply_filter({"started__year__gt": "2015"}, [self.data_1, self.data_2])
        self._apply_filter(
            {"started__year__gt": "2015", "started__month__gte": "8"}, [self.data_2]
        )

    def test_filter_process(self):
        self._apply_filter({"process": self.proc_1.pk}, [self.data_1])
        self._apply_filter({"process__slug": "test-process-1"}, [self.data_1])

    def test_filter_tags(self):
        self._apply_filter({"tags": "foo"}, [self.data_2])
        self._apply_filter({"tags": "bar"}, [self.data_2, self.data_3])
        self._apply_filter({"tags": "bar,moo"}, [self.data_2])


class CollectionFilterTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user_1 = get_user_model().objects.create(username="first_user")
        cls.user_2 = get_user_model().objects.create(username="second_user")

        tzone = get_current_timezone()

        cls.descriptor_schema_1 = DescriptorSchema.objects.create(
            contributor=cls.user_1, name="Descriptor schema 1"
        )

        cls.descriptor_schema_2 = DescriptorSchema.objects.create(
            contributor=cls.user_2, name="Descriptor schema 2"
        )

        cls.collection_1 = Collection.objects.create(
            name="Test collection 1",
            slug="test-collection-1",
            descriptor_schema=cls.descriptor_schema_1,
            contributor=cls.user_1,
        )
        cls.collection_1.created = datetime.datetime(2016, 7, 30, 13, 59, tzinfo=tzone)
        cls.collection_1.save()

        cls.collection_2 = Collection.objects.create(
            name="Test collection 2",
            slug="test-collection-2",
            descriptor_schema=cls.descriptor_schema_1,
            contributor=cls.user_1,
        )
        cls.collection_2.created = datetime.datetime(2016, 8, 30, 14, 59, tzinfo=tzone)
        cls.collection_2.save()

        cls.collection_3 = Collection.objects.create(
            name="Another collection",
            slug="another-collection",
            descriptor_schema=cls.descriptor_schema_2,
            contributor=cls.user_2,
        )
        cls.collection_3.created = datetime.datetime(2013, 1, 15, 7, 59, tzinfo=tzone)
        cls.collection_3.save()

        cls.proc = Process.objects.create(
            name="Test process",
            slug="test-process",
            contributor=cls.user_1,
            type="data:test:",
        )

        cls.data = Data.objects.create(
            name="Test data",
            slug="test-data",
            contributor=cls.user_1,
            process=cls.proc,
            status=Data.STATUS_DONE,
        )
        cls.collection_1.data.add(cls.data)

    def _apply_filter(self, filters, expected):
        filtered = CollectionFilter(filters, queryset=Collection.objects.all())
        self.assertCountEqual(filtered.qs, expected)

    def test_filter_id(self):
        self._apply_filter({"id": self.collection_1.pk}, [self.collection_1])
        self._apply_filter(
            {"id__in": "{},{}".format(self.collection_1.pk, self.collection_2.pk)},
            [self.collection_1, self.collection_2],
        )

    def test_filter_slug(self):
        self._apply_filter({"slug": "test-collection-2"}, [self.collection_2])
        self._apply_filter(
            {"slug__startswith": "test"}, [self.collection_1, self.collection_2]
        )

    def test_filter_name(self):
        self._apply_filter({"name": "Test collection 2"}, [self.collection_2])
        self._apply_filter(
            {"name__startswith": "Test"}, [self.collection_1, self.collection_2]
        )

    def test_filter_contributor(self):
        self._apply_filter(
            {"contributor": self.user_1.pk}, [self.collection_1, self.collection_2]
        )

    def test_filter_created(self):
        self._apply_filter(
            {"created__year": "2016"}, [self.collection_1, self.collection_2]
        )
        self._apply_filter(
            {"created__year__gt": "2015"}, [self.collection_1, self.collection_2]
        )
        self._apply_filter(
            {"created__year__gt": "2015", "created__month__gte": "8"},
            [self.collection_2],
        )

    def test_filter_modified(self):
        year = datetime.date.today().strftime("%Y")
        self._apply_filter(
            {"modified__year": year},
            [self.collection_1, self.collection_2, self.collection_3],
        )
        self._apply_filter({"modified__year__gt": year}, [])

    def test_filter_data(self):
        self._apply_filter({"data": self.data.pk}, [self.collection_1])

    def test_filter_descriptor_schema(self):
        self._apply_filter(
            {"descriptor_schema": self.descriptor_schema_1.pk},
            [self.collection_1, self.collection_2],
        )


class EntityFilterTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user_1 = get_user_model().objects.create(username="first_user")

        cls.collection_1 = Collection.objects.create(contributor=cls.user_1)
        cls.collection_2 = Collection.objects.create(contributor=cls.user_1)

        cls.entity_1 = Entity.objects.create(
            contributor=cls.user_1, collection=cls.collection_1
        )
        cls.entity_2 = Entity.objects.create(
            contributor=cls.user_1, collection=cls.collection_2
        )

    def _apply_filter(self, filters, expected):
        filtered = EntityFilter(filters, queryset=Entity.objects.all())
        self.assertCountEqual(filtered.qs, expected)

    def test_collection(self):
        self._apply_filter({"collection": self.collection_1.id}, [self.entity_1])
        self._apply_filter({"collection": self.collection_2.id}, [self.entity_2])



""".. Ignore pydocstyle D400.

.. autoclass:: resolwe.flow.executors.docker.run.FlowExecutor
    :members:

"""
# pylint: disable=logging-format-interpolation
import asyncio
import json
import logging
import os
import shlex
import tempfile
import time
from asyncio import subprocess

from ..global_settings import DATA_LOCATION, PROCESS_META, SETTINGS
from ..local.run import FlowExecutor as LocalFlowExecutor
from ..protocol import ExecutorFiles
from . import constants
from .seccomp import SECCOMP_POLICY

DOCKER_START_TIMEOUT = 60

# Limits of containers' access to memory. We set the limit to ensure
# processes are stable and do not get killed by OOM signal.
DOCKER_MEMORY_HARD_LIMIT_BUFFER = 100
DOCKER_MEMORY_SWAP_RATIO = 2
DOCKER_MEMORY_SWAPPINESS = 1

logger = logging.getLogger(__name__)


class FlowExecutor(LocalFlowExecutor):
    """Docker executor."""

    name = "docker"

    def __init__(self, *args, **kwargs):
        """Initialize attributes."""
        super().__init__(*args, **kwargs)

        self.container_name_prefix = None
        self.tools_volumes = None
        self.temporary_files = []
        self.command = SETTINGS.get("FLOW_DOCKER_COMMAND", "docker")

    def _generate_container_name(self):
        """Generate unique container name."""
        return "{}_{}".format(self.container_name_prefix, self.data_id)

    async def start(self):
        """Start process execution."""
        # arguments passed to the Docker command
        command_args = {
            "command": self.command,
            "container_image": self.requirements.get(
                "image", constants.DEFAULT_CONTAINER_IMAGE
            ),
        }

        # Get limit defaults.
        limit_defaults = SETTINGS.get("FLOW_PROCESS_RESOURCE_DEFAULTS", {})

        # Set resource limits.
        limits = []
        # Specify CPU limits. Default docker container is unlimited (0.000)
        # Available in Docker 1.13 and higher.
        limits.append("--cpus={:.1f}".format(self.process["resource_limits"]["cores"]))

        # Some SWAP is needed to avoid OOM signal. Swappiness is low to prevent
        # extensive usage of SWAP (this would reduce the performance).
        memory = (
            self.process["resource_limits"]["memory"] + DOCKER_MEMORY_HARD_LIMIT_BUFFER
        )
        memory_swap = int(memory * DOCKER_MEMORY_SWAP_RATIO)

        limits.append("--memory={}m".format(memory))
        limits.append("--memory-swap={}m".format(memory_swap))
        limits.append(
            "--memory-reservation={}m".format(self.process["resource_limits"]["memory"])
        )
        limits.append("--memory-swappiness={}".format(DOCKER_MEMORY_SWAPPINESS))

        # Set ulimits for interactive processes to prevent them from running too long.
        if (
            self.process["scheduling_class"]
            == PROCESS_META["SCHEDULING_CLASS_INTERACTIVE"]
        ):
            # TODO: This is not very good as each child gets the same limit.
            # Note: Ulimit does not work as expected on multithreaded processes
            # Limit is increased by factor 1.2 for processes with 2-8 threads.
            # TODO: This should be changed for processes with over 8 threads.
            cpu_time_interactive = limit_defaults.get("cpu_time_interactive", 30)
            limits.append("--ulimit cpu={}".format(int(cpu_time_interactive * 1.2)))

        command_args["limits"] = " ".join(limits)

        # set container name
        self.container_name_prefix = SETTINGS.get("FLOW_EXECUTOR", {}).get(
            "CONTAINER_NAME_PREFIX", "resolwe"
        )
        command_args["container_name"] = "--name={}".format(
            self._generate_container_name()
        )

        if "network" in self.resources:
            # Configure Docker network mode for the container (if specified).
            # By default, current Docker versions use the 'bridge' mode which
            # creates a network stack on the default Docker bridge.
            network = SETTINGS.get("FLOW_EXECUTOR", {}).get("NETWORK", "")
            command_args["network"] = "--net={}".format(network) if network else ""
        else:
            # No network if not specified.
            command_args["network"] = "--net=none"

        # Security options.
        security = []

        # Generate and set seccomp policy to limit syscalls.
        policy_file = tempfile.NamedTemporaryFile(mode="w")
        json.dump(SECCOMP_POLICY, policy_file)
        policy_file.file.flush()
        if not SETTINGS.get("FLOW_DOCKER_DISABLE_SECCOMP", False):
            security.append("--security-opt seccomp={}".format(policy_file.name))
        self.temporary_files.append(policy_file)

        # Drop all capabilities and only add ones that are needed.
        security.append("--cap-drop=all")

        command_args["security"] = " ".join(security)

        # Setup Docker volumes.
        def new_volume(kind, base_dir_name, volume, path=None, read_only=True):
            """Generate a new volume entry.

            :param kind: Kind of volume, which is used for getting extra options from
                settings (the ``FLOW_DOCKER_VOLUME_EXTRA_OPTIONS`` setting)
            :param base_dir_name: Name of base directory setting for volume source path
            :param volume: Destination volume mount point
            :param path: Optional additional path atoms appended to source path
            :param read_only: True to make the volume read-only
            """
            if path is None:
                path = []

            path = [str(atom) for atom in path]

            options = set(
                SETTINGS.get("FLOW_DOCKER_VOLUME_EXTRA_OPTIONS", {})
                .get(kind, "")
                .split(",")
            )
            options.discard("")
            # Do not allow modification of read-only option.
            options.discard("ro")
            options.discard("rw")

            if read_only:
                options.add("ro")
            else:
                options.add("rw")

            return {
                "src": os.path.join(
                    SETTINGS["FLOW_EXECUTOR"].get(base_dir_name, ""), *path
                ),
                "dest": volume,
                "options": ",".join(options),
            }

        volumes = [
            new_volume(
                "data",
                "DATA_DIR",
                constants.DATA_VOLUME,
                [DATA_LOCATION["subpath"]],
                read_only=False,
            ),
            new_volume("data_all", "DATA_DIR", constants.DATA_ALL_VOLUME),
            new_volume(
                "upload", "UPLOAD_DIR", constants.UPLOAD_VOLUME, read_only=False
            ),
            new_volume(
                "secrets",
                "RUNTIME_DIR",
                constants.SECRETS_VOLUME,
                [DATA_LOCATION["subpath"], ExecutorFiles.SECRETS_DIR],
            ),
        ]

        # Generate dummy passwd and create mappings for it. This is required because some tools
        # inside the container may try to lookup the given UID/GID and will crash if they don't
        # exist. So we create minimal user/group files.
        passwd_file = tempfile.NamedTemporaryFile(mode="w")
        passwd_file.write("root:x:0:0:root:/root:/bin/bash\n")
        passwd_file.write(
            "user:x:{}:{}:user:/:/bin/bash\n".format(os.getuid(), os.getgid())
        )
        passwd_file.file.flush()
        self.temporary_files.append(passwd_file)

        group_file = tempfile.NamedTemporaryFile(mode="w")
        group_file.write("root:x:0:\n")
        group_file.write("user:x:{}:user\n".format(os.getgid()))
        group_file.file.flush()
        self.temporary_files.append(group_file)

        volumes += [
            new_volume("users", None, "/etc/passwd", [passwd_file.name]),
            new_volume("users", None, "/etc/group", [group_file.name]),
        ]

        # Create volumes for tools.
        # NOTE: To prevent processes tampering with tools, all tools are mounted read-only
        self.tools_volumes = []
        for index, tool in enumerate(self.get_tools_paths()):
            self.tools_volumes.append(
                new_volume(
                    "tools",
                    None,
                    os.path.join("/usr/local/bin/resolwe", str(index)),
                    [tool],
                )
            )

        volumes += self.tools_volumes

        # Create volumes for runtime (all read-only).
        runtime_volume_maps = SETTINGS.get("RUNTIME_VOLUME_MAPS", None)
        if runtime_volume_maps:
            for src, dst in runtime_volume_maps.items():
                volumes.append(
                    new_volume(
                        "runtime", "RUNTIME_DIR", dst, [DATA_LOCATION["subpath"], src],
                    )
                )

        # Add any extra volumes verbatim.
        volumes += SETTINGS.get("FLOW_DOCKER_EXTRA_VOLUMES", [])

        # Make sure that tmp dir exists.
        os.makedirs(constants.TMPDIR, mode=0o755, exist_ok=True)

        # Create Docker --volume parameters from volumes.
        command_args["volumes"] = " ".join(
            [
                '--volume="{src}":"{dest}":{options}'.format(**volume)
                for volume in volumes
            ]
        )

        # Set working directory to the data volume.
        command_args["workdir"] = "--workdir={}".format(constants.DATA_VOLUME)

        # Change user inside the container.
        command_args["user"] = "--user={}:{}".format(os.getuid(), os.getgid())

        # A non-login Bash shell should be used here (a subshell will be spawned later).
        command_args["shell"] = "/bin/bash"

        # Check if image exists locally. If not, command will exit with non-zero returncode
        check_command = "{command} image inspect {container_image}".format(
            **command_args
        )

        logger.debug(
            "Checking existence of docker image: {}".format(
                command_args["container_image"]
            )
        )

        check_proc = await subprocess.create_subprocess_exec(
            *shlex.split(check_command), stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        await check_proc.communicate()

        if check_proc.returncode != 0:
            pull_command = "{command} pull {container_image}".format(**command_args)

            logger.info(
                "Pulling docker image: {}".format(command_args["container_image"])
            )

            pull_proc = await subprocess.create_subprocess_exec(
                *shlex.split(pull_command),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            _, stderr = await pull_proc.communicate()

            if pull_proc.returncode != 0:
                error_msg = "Docker failed to pull {} image.".format(
                    command_args["container_image"]
                )
                if stderr:
                    error_msg = "\n".join([error_msg, stderr.decode("utf-8")])
                raise RuntimeError(error_msg)

        docker_command = (
            "{command} run --rm --interactive {container_name} {network} {volumes} {limits} "
            "{security} {workdir} {user} {container_image} {shell}".format(
                **command_args
            )
        )

        logger.info("Starting docker container with command: {}".format(docker_command))
        start_time = time.time()

        # Workaround for pylint issue #1469
        # (https://github.com/PyCQA/pylint/issues/1469).
        self.proc = await subprocess.create_subprocess_exec(
            *shlex.split(docker_command),
            limit=4 * (2 ** 20),  # 4MB buffer size for line buffering
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

        stdout = []

        async def wait_for_container():
            """Wait for Docker container to start to avoid blocking the code that uses it."""
            self.proc.stdin.write(("echo PING" + os.linesep).encode("utf-8"))
            await self.proc.stdin.drain()
            while True:
                line = await self.proc.stdout.readline()
                stdout.append(line)
                if line.rstrip() == b"PING":
                    break
                if self.proc.stdout.at_eof():
                    raise RuntimeError()

        try:
            await asyncio.wait_for(wait_for_container(), timeout=DOCKER_START_TIMEOUT)
        except (asyncio.TimeoutError, RuntimeError):
            error_msg = "Docker container has not started for {} seconds.".format(
                DOCKER_START_TIMEOUT
            )
            stdout = "".join([line.decode("utf-8") for line in stdout if line])
            if stdout:
                error_msg = "\n".join([error_msg, stdout])
            raise RuntimeError(error_msg)

        end_time = time.time()
        logger.info(
            "It took {:.2f}s for Docker container to start".format(
                end_time - start_time
            )
        )

        self.stdout = self.proc.stdout

    async def run_script(self, script):
        """Execute the script and save results."""
        # Create a Bash command to add all the tools to PATH.
        tools_paths = ":".join([map_["dest"] for map_ in self.tools_volumes])
        add_tools_path = "export PATH=$PATH:{}".format(tools_paths)
        # Spawn another child bash, to avoid running anything as PID 1, which has special
        # signal handling (e.g., cannot be SIGKILL-ed from inside).
        # A login Bash shell is needed to source /etc/profile.
        bash_line = "/bin/bash --login; exit $?" + os.linesep
        script = (
            os.linesep.join(["set -x", "set +B", add_tools_path, script]) + os.linesep
        )
        self.proc.stdin.write(bash_line.encode("utf-8"))
        await self.proc.stdin.drain()
        self.proc.stdin.write(script.encode("utf-8"))
        await self.proc.stdin.drain()
        self.proc.stdin.close()

    async def end(self):
        """End process execution."""
        try:
            await self.proc.wait()
        finally:
            # Cleanup temporary files.
            for temporary_file in self.temporary_files:
                temporary_file.close()
            self.temporary_files = []

        return self.proc.returncode

    async def terminate(self):
        """Terminate a running script."""
        # Workaround for pylint issue #1469
        # (https://github.com/PyCQA/pylint/issues/1469).
        cmd = await subprocess.create_subprocess_exec(
            *shlex.split(
                "{} rm -f {}".format(self.command, self._generate_container_name())
            )
        )
        await cmd.wait()
        await self.proc.wait()

        await super().terminate()



from django.core.exceptions import ValidationError
from django.db import migrations

from django.utils.timezone import now

from resolwe.flow.utils import rewire_inputs


def copy_permissions(original, duplicate, model, apps):
    """Copy permissions from original object to it's duplicate."""
    UserObjectPermission = apps.get_model("guardian", "UserObjectPermission")
    GroupObjectPermission = apps.get_model("guardian", "GroupObjectPermission")

    kwargs = {
        "permission__content_type__model": model,
        "object_pk": original.id,
    }
    for perm in UserObjectPermission.objects.filter(**kwargs):
        perm.id = None
        perm.object_pk = duplicate.id
        perm.content_object = duplicate
        perm.save(force_insert=True)

    for perm in GroupObjectPermission.objects.filter(**kwargs):
        perm.id = None
        perm.object_pk = duplicate.id
        perm.content_object = duplicate
        perm.save(force_insert=True)


def duplicate_entity(original, apps):
    """Duplicate entity."""
    Entity = apps.get_model("flow", "Entity")

    duplicate = Entity.objects.get(id=original.id)
    duplicate.pk = None
    duplicate.slug = None
    duplicate.collection = None
    duplicate.name = "Copy of {}".format(original.name)
    duplicate.duplicated = now()

    duplicate.save(force_insert=True)

    copy_permissions(original, duplicate, model="entity", apps=apps)

    # Override fields that are automatically set on create.
    duplicate.created = original.created
    duplicate.save()

    # Duplicate entity's data objects.
    bundle = [
        {"original": obj, "copy": duplicate_data(obj, apps)}
        for obj in original.data.all()
    ]
    bundle = rewire_inputs(bundle)
    duplicated_data = [item["copy"] for item in bundle]

    duplicate.data.add(*duplicated_data)

    return duplicate


def duplicate_data(original, apps):
    """Duplicate (make a copy)."""
    DataDependency = apps.get_model("flow", "DataDependency")
    Data = apps.get_model("flow", "Data")

    if original.status not in ["OK", "ER"]:
        raise ValidationError(
            "Data object must have done or error status to be duplicated"
        )

    duplicate = Data.objects.get(id=original.id)
    duplicate.pk = None
    duplicate.slug = None
    duplicate.entity2 = None
    duplicate.collection = None
    duplicate.name = "Copy of {}".format(original.name)
    duplicate.duplicated = now()

    duplicate.save(force_insert=True)

    copy_permissions(original, duplicate, model="data", apps=apps)

    # Override fields that are automatically set on create.
    duplicate.created = original.created
    duplicate.save()

    if original.location:
        original.location.data.add(duplicate)

    duplicate.storages.set(original.storages.all())

    for migration in original.migration_history.order_by("created"):
        migration.pk = None
        migration.data = duplicate
        migration.save(force_insert=True)

    # Inherit existing child dependencies.
    DataDependency.objects.bulk_create(
        [
            DataDependency(
                child=duplicate, parent=dependency.parent, kind=dependency.kind
            )
            for dependency in DataDependency.objects.filter(child=original)
        ]
    )
    # Inherit existing parent dependencies.
    DataDependency.objects.bulk_create(
        [
            DataDependency(
                child=dependency.child, parent=duplicate, kind=dependency.kind
            )
            for dependency in DataDependency.objects.filter(parent=original)
        ]
    )

    return duplicate


def main(apps, schema_editor):
    """Reassign Data, Entities and Collections relations from m2m to m2o fields."""
    Data = apps.get_model("flow", "Data")
    Entity = apps.get_model("flow", "Entity")

    # 1: Migrate entities from multiple collections.
    for entity in Entity.objects.iterator():

        count = entity.collections.count()
        if count == 1:
            entity.collection = entity.collections.first()
            entity.save()
        elif count > 1:
            collections = entity.collections.all()
            for i, collection in enumerate(collections, start=1):
                # For the last collection, we can keep existing one.
                if i == count:
                    duplicate = entity
                else:
                    duplicate = duplicate_entity(entity, apps)

                duplicate.collection = collection
                duplicate.save()
                for datum in duplicate.data.all():
                    datum.collection = collection
                    datum.save()

    # 2: Migrate data from multiple collections.
    for data in Data.objects.iterator():

        count = data.collection_set.count()
        if count == 1:
            data.collection = data.collection_set.first()
            data.save()
        elif count > 1:
            collections = data.collection_set.all()
            for i, collection in enumerate(collections, start=1):
                # For the last collection, we can keep existing one.
                if i == count:
                    duplicate = data
                else:
                    duplicate = duplicate_data(data, apps)

                duplicate.collection = collection
                duplicate.save()

    # 3: Migrate data from multiple entities.
    for data in Data.objects.iterator():

        count = data.entity_set.count()
        if count == 1:
            data.entity2 = data.entity_set.first()
            data.save()
        elif count > 1:
            entities = data.entity_set.all()
            for i, entity in enumerate(entities, start=1):
                # For the last entity, we can keep existing one.
                if i == count:
                    duplicate = data
                else:
                    duplicate = duplicate_data(data, apps)

                duplicate.entity2 = entity
                duplicate.save()


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0036_add_m2o_fields"),
        ("contenttypes", "0002_remove_content_type_name"),
        ("guardian", "0001_initial"),
    ]

    operations = [
        migrations.RunPython(main, atomic=True),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-21 10:43
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0003_data_dependency_1"),
    ]

    operations = [
        migrations.AddField(
            model_name="datadependency",
            name="kind",
            field=models.CharField(
                choices=[
                    ("io", "Input/output dependency"),
                    ("subprocess", "Subprocess"),
                ],
                default="io",
                max_length=16,
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="datadependency",
            name="child",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="parents_dependency",
                to="flow.Data",
            ),
        ),
        migrations.AlterField(
            model_name="datadependency",
            name="parent",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="children_dependency",
                to="flow.Data",
            ),
        ),
        migrations.AlterModelTable(name="datadependency", table=None,),
    ]



# pylint: disable=missing-docstring
from resolwe.elastic.composer import composer
from resolwe.test import TestCase


class MyExtendableType:
    pass


class ComposerTest(TestCase):
    def test_add_extension(self):
        # Add some extensions.
        composer.add_extension(
            "resolwe.elastic.tests.test_composer.MyExtendableType", 42
        )
        composer.add_extension(
            "resolwe.elastic.tests.test_composer.MyExtendableType", "foo"
        )

        # Check that they are all here.
        extensions = composer.get_extensions(MyExtendableType)
        self.assertEqual(extensions, [42, "foo"])

        extensions = composer.get_extensions(
            "resolwe.elastic.tests.test_composer.MyExtendableType"
        )
        self.assertEqual(extensions, [42, "foo"])

        extensions = composer.get_extensions(MyExtendableType())
        self.assertEqual(extensions, [42, "foo"])



""".. Ignore pydocstyle D400.

======================
Flow Executor Preparer
======================

Framework for the manager-resident executor preparation facilities.

.. autoclass:: resolwe.flow.executors.prepare.BaseFlowExecutorPreparer
    :members:

"""
import logging
import os

from django.conf import settings
from django.forms.models import model_to_dict

from resolwe.flow.managers.protocol import ExecutorFiles
from resolwe.flow.models import Data
from resolwe.flow.utils import get_apps_tools
from resolwe.test.utils import is_testing

logger = logging.getLogger(__name__)


class BaseFlowExecutorPreparer:
    """Represents the preparation functionality of the executor."""

    def extend_settings(self, data_id, files, secrets):
        """Extend the settings the manager will serialize.

        :param data_id: The :class:`~resolwe.flow.models.Data` object id
            being prepared for.
        :param files: The settings dictionary to be serialized. Keys are
            filenames, values are the objects that will be serialized
            into those files. Standard filenames are listed in
            ``resolwe.flow.managers.protocol.ExecutorFiles``.
        :param secrets: Secret files dictionary describing additional secret
            file content that should be created and made available to
            processes with special permissions. Keys are filenames, values
            are the raw strings that should be written into those files.
        """
        data = Data.objects.select_related("process").get(pk=data_id)

        files[ExecutorFiles.DJANGO_SETTINGS].update(
            {
                "USE_TZ": settings.USE_TZ,
                "FLOW_EXECUTOR_TOOLS_PATHS": self.get_tools_paths(),
            }
        )
        files[ExecutorFiles.DATA] = model_to_dict(data)
        files[ExecutorFiles.DATA_LOCATION] = model_to_dict(data.location)
        files[ExecutorFiles.PROCESS] = model_to_dict(data.process)
        files[ExecutorFiles.PROCESS][
            "resource_limits"
        ] = data.process.get_resource_limits()

        # Add secrets if the process has permission to read them.
        secrets.update(data.resolve_secrets())

    def get_tools_paths(self):
        """Get tools' paths."""
        if settings.DEBUG or is_testing():
            return list(get_apps_tools().values())

        else:
            tools_root = settings.FLOW_TOOLS_ROOT
            subdirs = next(os.walk(tools_root))[1]

            return [os.path.join(tools_root, sdir) for sdir in subdirs]

    def post_register_hook(self, verbosity=1):
        """Run hook after the 'register' management command finishes.

        Subclasses may implement this hook to e.g. pull Docker images at
        this point. By default, it does nothing.
        """

    def resolve_data_path(self, data=None, filename=None):
        """Resolve data path for use with the executor.

        :param data: Data object instance
        :param filename: Filename to resolve
        :return: Resolved filename, which can be used to access the
            given data file in programs executed using this executor
        """
        if data is None:
            return settings.FLOW_EXECUTOR["DATA_DIR"]

        return data.location.get_path(filename=filename)

    def resolve_upload_path(self, filename=None):
        """Resolve upload path for use with the executor.

        :param filename: Filename to resolve
        :return: Resolved filename, which can be used to access the
            given uploaded file in programs executed using this
            executor
        """
        if filename is None:
            return settings.FLOW_EXECUTOR["UPLOAD_DIR"]

        return os.path.join(settings.FLOW_EXECUTOR["UPLOAD_DIR"], filename)

    def get_environment_variables(self):
        """Return dict of environment variables that will be added to executor."""
        return {}



"""Resolwe data serializer."""
from rest_framework import serializers

from resolwe.flow.models import Collection, Data, DescriptorSchema, Entity, Process
from resolwe.rest.fields import ProjectableJSONField

from .base import ResolweBaseSerializer
from .collection import CollectionSerializer
from .descriptor import DescriptorSchemaSerializer
from .entity import EntitySerializer
from .fields import DictRelatedField
from .process import ProcessSerializer


class DataSerializer(ResolweBaseSerializer):
    """Serializer for Data objects."""

    input = ProjectableJSONField(required=False)
    output = ProjectableJSONField(required=False)
    descriptor = ProjectableJSONField(required=False)
    process = DictRelatedField(
        queryset=Process.objects.all(), serializer=ProcessSerializer
    )
    descriptor_schema = DictRelatedField(
        queryset=DescriptorSchema.objects.all(),
        serializer=DescriptorSchemaSerializer,
        allow_null=True,
        required=False,
    )
    collection = DictRelatedField(
        queryset=Collection.objects.all(),
        serializer=CollectionSerializer,
        allow_null=True,
        required=False,
        write_permission="edit",
    )
    entity = DictRelatedField(
        queryset=Entity.objects.all(),
        serializer=EntitySerializer,
        allow_null=True,
        required=False,
        write_permission="edit",
    )

    class Meta:
        """DataSerializer Meta options."""

        model = Data
        read_only_fields = (
            "checksum",
            "created",
            "descriptor_dirty",
            "duplicated",
            "finished",
            "id",
            "modified",
            "process_cores",
            "process_error",
            "process_info",
            "process_memory",
            "process_progress",
            "process_rc",
            "process_warning",
            "output",
            "scheduled",
            "size",
            "started",
            "status",
        )
        update_protected_fields = (
            "contributor",
            "input",
            "process",
        )
        fields = (
            read_only_fields
            + update_protected_fields
            + (
                "collection",
                "descriptor",
                "descriptor_schema",
                "entity",
                "name",
                "slug",
                "tags",
            )
        )

    def validate_process(self, process):
        """Check that process is active."""
        if not process.is_active:
            raise serializers.ValidationError(
                "Process {} is not active.".format(process)
            )
        return process

    def validate_collection(self, collection):
        """Verify that changing collection is done in the right place."""
        if getattr(self.instance, "entity", None):
            if getattr(self.instance.entity.collection, "id", None) != getattr(
                collection, "id", None
            ):
                raise serializers.ValidationError(
                    "If Data is in entity, you can only move it to another collection "
                    "by moving entire entity."
                )
        return collection



""".. Ignore pydocstyle D400.

==========
Base Class
==========

.. autoclass:: resolwe.flow.executors.run.BaseFlowExecutor
    :members:

"""
# pylint: disable=logging-format-interpolation
import asyncio
import json
import logging
import os
import shutil
import signal
import sys
import uuid
from collections import defaultdict

from .global_settings import DATA_META, EXECUTOR_SETTINGS, PROCESS, SETTINGS
from .manager_commands import send_manager_command
from .protocol import ExecutorProtocol

# NOTE: If the imports here are changed, the executors' requirements.txt
# file must also be updated accordingly.

logger = logging.getLogger(__name__)


def iterjson(text):
    """Decode JSON stream."""
    decoder = json.JSONDecoder()
    while text:
        obj, ndx = decoder.raw_decode(text)

        if not isinstance(obj, dict):
            raise ValueError()

        text = text[ndx:].lstrip("\r\n")
        yield obj


class BaseFlowExecutor:
    """Represents a workflow executor."""

    exported_files_mapper = defaultdict(dict)

    def __init__(self, *args, **kwargs):
        """Initialize attributes."""
        self.data_id = None
        self.process = None
        self.requirements = {}
        self.resources = {}

        asyncio.get_event_loop().add_signal_handler(
            signal.SIGTERM, lambda: asyncio.ensure_future(self._exit_gracefully())
        )

    async def _exit_gracefully(self):
        """Handle SIGTERM signal."""
        await self.update_data_status(
            process_error=["Executor was killed by the scheduling system."],
            status=DATA_META["STATUS_ERROR"],
        )

        await self.terminate()

    async def _send_manager_command(self, *args, **kwargs):
        """Send an update to manager and terminate the process if it fails."""
        resp = await send_manager_command(*args, **kwargs)

        if resp is False:
            await self.terminate()

    def get_tools_paths(self):
        """Get tools paths."""
        tools_paths = SETTINGS["FLOW_EXECUTOR_TOOLS_PATHS"]
        return tools_paths

    async def start(self):
        """Start process execution."""

    async def run_script(self, script):
        """Run process script."""
        raise NotImplementedError(
            "Subclasses of BaseFlowExecutor must implement a run_script() method."
        )

    async def end(self):
        """End process execution."""

    def get_stdout(self):
        """Get process' standard output."""
        return self.stdout

    async def update_data_status(self, **kwargs):
        """Update (PATCH) Data object.

        :param kwargs: The dictionary of
            :class:`~resolwe.flow.models.Data` attributes to be changed.
        """
        await self._send_manager_command(
            ExecutorProtocol.UPDATE,
            extra_fields={ExecutorProtocol.UPDATE_CHANGESET: kwargs},
        )

    async def run(self, data_id, script):
        """Execute the script and save results."""
        logger.debug("Executor for Data with id {} has started.".format(data_id))
        try:
            finish_fields = await self._run(data_id, script)
        except SystemExit as ex:
            raise ex
        except Exception as error:
            logger.exception("Unhandled exception in executor")

            # Send error report.
            await self.update_data_status(
                process_error=[str(error)], status=DATA_META["STATUS_ERROR"]
            )

            finish_fields = {
                ExecutorProtocol.FINISH_PROCESS_RC: 1,
            }

        if finish_fields is not None:
            await self._send_manager_command(
                ExecutorProtocol.FINISH, extra_fields=finish_fields
            )

        # The response channel (Redis list) is deleted automatically once the list is drained, so
        # there is no need to remove it manually.

    def _create_file(self, filename):
        """Ensure a new file is created and opened for writing."""
        file_descriptor = os.open(filename, os.O_WRONLY | os.O_CREAT | os.O_EXCL)
        return os.fdopen(file_descriptor, "w")

    async def _run(self, data_id, script):
        """Execute the script and save results."""
        self.data_id = data_id

        # Fetch data instance to get any executor requirements.
        self.process = PROCESS
        requirements = self.process["requirements"]
        self.requirements = requirements.get("executor", {}).get(self.name, {})
        self.resources = requirements.get("resources", {})

        logger.debug("Preparing output files for Data with id {}".format(data_id))
        os.chdir(EXECUTOR_SETTINGS["DATA_DIR"])
        try:
            log_file = self._create_file("stdout.txt")
            json_file = self._create_file("jsonout.txt")
        except FileExistsError:
            logger.error("Stdout or jsonout out file already exists.")
            # Looks like executor was already ran for this Data object,
            # so don't raise the error to prevent setting status to error.
            await self._send_manager_command(ExecutorProtocol.ABORT, expect_reply=False)
            return

        proc_pid = await self.start()

        await self.update_data_status(
            status=DATA_META["STATUS_PROCESSING"], process_pid=proc_pid
        )

        # Run process and handle intermediate results
        logger.info("Running program for Data with id {}".format(data_id))
        logger.debug("The program for Data with id {} is: \n{}".format(data_id, script))
        await self.run_script(script)
        spawn_processes = []
        output = {}
        process_error, process_warning, process_info = [], [], []
        process_progress, process_rc = 0, 0

        # read process output
        try:
            stdout = self.get_stdout()
            while True:
                line = await stdout.readline()
                logger.debug("Process's output: {}".format(line.strip()))

                if not line:
                    break
                line = line.decode("utf-8")

                try:
                    if line.strip().startswith("run"):
                        # Save process and spawn if no errors
                        log_file.write(line)
                        log_file.flush()

                        for obj in iterjson(line[3:].strip()):
                            spawn_processes.append(obj)
                    elif line.strip().startswith("export"):
                        file_name = line[6:].strip()

                        export_folder = SETTINGS["FLOW_EXECUTOR"]["UPLOAD_DIR"]
                        unique_name = "export_{}".format(uuid.uuid4().hex)
                        export_path = os.path.join(export_folder, unique_name)

                        self.exported_files_mapper[self.data_id][
                            file_name
                        ] = unique_name

                        shutil.move(file_name, export_path)
                    else:
                        # If JSON, save to MongoDB
                        updates = {}
                        for obj in iterjson(line):
                            for key, val in obj.items():
                                if key.startswith("proc."):
                                    if key == "proc.error":
                                        process_error.append(val)
                                        if not process_rc:
                                            process_rc = 1
                                            updates["process_rc"] = process_rc
                                        updates["process_error"] = process_error
                                        updates["status"] = DATA_META["STATUS_ERROR"]
                                    elif key == "proc.warning":
                                        process_warning.append(val)
                                        updates["process_warning"] = process_warning
                                    elif key == "proc.info":
                                        process_info.append(val)
                                        updates["process_info"] = process_info
                                    elif key == "proc.rc":
                                        process_rc = int(val)
                                        updates["process_rc"] = process_rc
                                        if process_rc != 0:
                                            updates["status"] = DATA_META[
                                                "STATUS_ERROR"
                                            ]
                                    elif key == "proc.progress":
                                        process_progress = int(float(val) * 100)
                                        updates["process_progress"] = process_progress
                                else:
                                    output[key] = val
                                    updates["output"] = output

                        if updates:
                            await self.update_data_status(**updates)
                            # Process meta fields are collected in listener, so we can clear them.
                            process_error, process_warning, process_info = [], [], []

                        if process_rc > 0:
                            log_file.close()
                            json_file.close()
                            await self._send_manager_command(
                                ExecutorProtocol.FINISH,
                                extra_fields={
                                    ExecutorProtocol.FINISH_PROCESS_RC: process_rc
                                },
                            )
                            return

                        # Debug output
                        # Not referenced in Data object
                        json_file.write(line)
                        json_file.flush()

                except ValueError:
                    # Ignore if not JSON
                    log_file.write(line)
                    log_file.flush()

        except MemoryError as ex:
            logger.error("Out of memory:\n\n{}".format(ex))

        except IOError as ex:
            # TODO: if ex.errno == 28: no more free space
            raise ex
        finally:
            # Store results
            log_file.close()
            json_file.close()

        return_code = await self.end()

        if process_rc < return_code:
            process_rc = return_code

        # send a notification to the executor listener that we're done
        finish_fields = {ExecutorProtocol.FINISH_PROCESS_RC: process_rc}
        if spawn_processes and process_rc == 0:
            finish_fields[ExecutorProtocol.FINISH_SPAWN_PROCESSES] = spawn_processes
            finish_fields[
                ExecutorProtocol.FINISH_EXPORTED_FILES
            ] = self.exported_files_mapper

        return finish_fields

    async def terminate(self):
        """Terminate a running script."""
        sys.exit(1)



"""Elastic search fields for Resolwe."""
import elasticsearch_dsl as dsl

# pylint: disable=invalid-name
# Process type analyzer. During indexing we tokenize by type paths, during search,
# we do not tokenize at all.
process_type_tokenizer = dsl.tokenizer(
    "process_type_tokenizer", type="path_hierarchy", delimiter=":"
)
process_type_analyzer = dsl.analyzer(
    "process_type_analyzer", tokenizer=process_type_tokenizer, filter=["lowercase"]
)
process_type_search_analyzer = dsl.analyzer(
    "process_type_search_analyzer", tokenizer="keyword", filter=["lowercase"]
)

# Name analyzer.
name_analyzer = dsl.analyzer(
    "name_analyzer",
    type="pattern",
    # The pattern matches token separators.
    pattern=r"""
          ([^\p{L}\d]+)                 # swallow non letters and numbers,
        | (?<=\D)(?=\d)                 # or non-number followed by number,
        | (?<=\d)(?=\D)                 # or number followed by non-number,
    """,
    flags="CASE_INSENSITIVE|COMMENTS",
    lowercase=True,
)

# During indexing, we lowercase terms and tokenize using edge_ngram.
ngrams_analyzer = dsl.analyzer(
    "ngrams_index",
    tokenizer="standard",
    filter=[
        "lowercase",
        dsl.token_filter("ngrams_filter", type="edgeNGram", min_gram=1, max_gram=15,),
    ],
)
# During search, we only lowercase terms.
ngrams_search_analyzer = dsl.analyzer(
    "ngrams_search", tokenizer="standard", filter=["lowercase"],
)
# pylint: enable=invalid-name


class RawKeywordSubfieldMixin:
    """String field with a 'raw' subfield (e.g. for sorting)."""

    def __init__(self, *args, **kwargs):
        """Construct field."""
        kwargs.setdefault("fields", {})["raw"] = {"type": "keyword"}
        super().__init__(*args, **kwargs)


class NgramsSubfieldMixin:
    """String field with a 'ngrams' subfield (e.g. for autocomplete)."""

    def __init__(self, *args, **kwargs):
        """Construct field."""
        kwargs.setdefault("fields", {})["ngrams"] = {
            "type": "text",
            "analyzer": ngrams_analyzer,
            "search_analyzer": ngrams_search_analyzer,
        }
        super().__init__(*args, **kwargs)


class Name(RawKeywordSubfieldMixin, NgramsSubfieldMixin, dsl.Text):
    """Field for names supporting term matches.

    Includes a 'raw' and 'ngrams' subfields for sorting and
    autocomplete.
    """

    def __init__(self, *args, **kwargs):
        """Construct field."""
        kwargs.setdefault("analyzer", name_analyzer)
        super().__init__(*args, **kwargs)


class ProcessType(RawKeywordSubfieldMixin, dsl.Text):
    """Field for process type supporting hierarchical type matches.

    Includes a 'raw' subfield for sorting.
    """

    def __init__(self, *args, **kwargs):
        """Construct field."""
        kwargs.setdefault("analyzer", process_type_analyzer)
        kwargs.setdefault("search_analyzer", process_type_search_analyzer)
        super().__init__(*args, **kwargs)


class User(NgramsSubfieldMixin, dsl.Text):
    """Field for users supporting partial matches."""


class Slug(NgramsSubfieldMixin, dsl.Keyword):
    """Field for slugs supporting partial matches."""



"""Resolwe relation serializer."""
from django.db import transaction

from rest_framework import serializers

from resolwe.flow.models.collection import Collection
from resolwe.flow.models.entity import Relation, RelationPartition, RelationType
from resolwe.rest.serializers import SelectiveFieldMixin

from .base import ResolweBaseSerializer


class RelationPartitionSerializer(SelectiveFieldMixin, serializers.ModelSerializer):
    """Serializer for RelationPartition objects."""

    class Meta:
        """RelationPartitionSerializer Meta options."""

        model = RelationPartition
        fields = ("id", "entity", "position", "label")


class RelationSerializer(ResolweBaseSerializer):
    """Serializer for Relation objects."""

    partitions = RelationPartitionSerializer(source="relationpartition_set", many=True)
    collection = serializers.PrimaryKeyRelatedField(queryset=Collection.objects.all())
    type = serializers.SlugRelatedField(
        queryset=RelationType.objects.all(), slug_field="name"
    )

    class Meta:
        """RelationSerializer Meta options."""

        model = Relation
        read_only_fields = (
            "created",
            "id",
            "modified",
        )
        update_protected_fields = (
            "contributor",
            "type",
        )
        fields = (
            read_only_fields
            + update_protected_fields
            + ("collection", "category", "partitions", "unit",)
        )

    def validate_partitions(self, partitions):
        """Raise validation error if list of partitions is empty."""
        if not partitions:
            raise serializers.ValidationError("List of partitions must not be empty.")

        return partitions

    def _create_partitions(self, instance, partitions):
        """Create partitions."""
        for partition in partitions:
            RelationPartition.objects.create(
                relation=instance,
                entity=partition["entity"],
                label=partition.get("label", None),
                position=partition.get("position", None),
            )

    def create(self, validated_data):
        """Create ``Relation`` object and add partitions of ``Entities``."""
        # `partitions` field is renamed to `relationpartition_set` based on source of nested serializer
        partitions = validated_data.pop("relationpartition_set")

        with transaction.atomic():
            instance = Relation.objects.create(**validated_data)
            self._create_partitions(instance, partitions)

        return instance

    def update(self, instance, validated_data):
        """Update ``Relation``."""
        # `partitions` field is renamed to `relationpartition_set` based on source of nested serializer
        partitions = validated_data.pop("relationpartition_set", None)

        with transaction.atomic():
            instance = super().update(instance, validated_data)

            if partitions is not None:
                # TODO: Apply the diff instead of recreating all objects.
                instance.relationpartition_set.all().delete()
                self._create_partitions(instance, partitions)

        return instance



"""Elastic search indexes for Resolwe models."""

from .collection import CollectionDocument, CollectionIndex
from .data import DataDocument, DataIndex
from .entity import EntityDocument, EntityIndex

__all__ = (
    "CollectionDocument",
    "CollectionIndex",
    "DataDocument",
    "DataIndex",
    "EntityDocument",
    "EntityIndex",
)



""".. Ignore pydocstyle D400.

=========================
Delete Unreferenced Files
=========================

"""
from django.core.management.base import BaseCommand

from resolwe.flow.utils.purge import purge_all


class Command(BaseCommand):
    """Purge files with no reference in Data objects, and orphaned storages."""

    help = "Purge files with no reference in Data objects, and orphaned storages."

    def add_arguments(self, parser):
        """Command arguments."""
        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            help="Delete unreferenced files and storages",
        )

    def handle(self, *args, **options):
        """Call :func:`~resolwe.flow.utils.purge.purge_all`."""
        purge_all(delete=options["force"], verbosity=options["verbosity"])



""".. Ignore pydocstyle D400.

==================
Permissions Filter
==================

"""
from rest_framework.filters import BaseFilterBackend

from resolwe.permissions.shortcuts import get_objects_for_user


class ResolwePermissionsFilter(BaseFilterBackend):
    """Permissions filter."""

    def filter_queryset(self, request, queryset, view):
        """Filter permissions queryset."""
        user = request.user

        app_label = queryset.model._meta.app_label
        model_name = queryset.model._meta.model_name

        kwargs = {}
        if model_name == "storage":
            model_name = "data"
            kwargs["perms_filter"] = "data__pk__in"

        if model_name == "relation":
            model_name = "collection"
            kwargs["perms_filter"] = "collection__pk__in"

        permission = "{}.view_{}".format(app_label, model_name)

        return get_objects_for_user(user, permission, queryset, **kwargs)



"""Utility functions for communicating with the manager."""
# pylint: disable=logging-format-interpolation
import json
import logging
import traceback

# NOTE: If the imports here are changed, the executors' requirements.txt
# file must also be updated accordingly.
import aioredis

from .global_settings import DATA, EXECUTOR_SETTINGS, SETTINGS
from .protocol import ExecutorProtocol

logger = logging.getLogger(__name__)

_REDIS_RETRIES = 60
# This channel name will be used for all listener communication; Data object-specific.
_response_channel = EXECUTOR_SETTINGS.get("REDIS_CHANNEL_PAIR", ("", ""))[1]
QUEUE_RESPONSE_CHANNEL = "{}.{}".format(_response_channel, DATA.get("id", 0))

# The Redis connection instance used to communicate with the manager listener.
redis_conn = None


async def init():
    """Create a connection to the Redis server."""
    global redis_conn
    conn = await aioredis.create_connection(
        "redis://{}:{}".format(
            SETTINGS.get("FLOW_EXECUTOR", {})
            .get("REDIS_CONNECTION", {})
            .get("host", "localhost"),
            SETTINGS.get("FLOW_EXECUTOR", {})
            .get("REDIS_CONNECTION", {})
            .get("port", 56379),
        ),
        db=int(
            SETTINGS.get("FLOW_EXECUTOR", {}).get("REDIS_CONNECTION", {}).get("db", 1)
        ),
    )
    redis_conn = aioredis.Redis(conn)


async def deinit():
    """Close the Redis connection cleanly."""
    redis_conn.close()
    await redis_conn.wait_closed()


async def send_manager_command(cmd, expect_reply=True, extra_fields={}):
    """Send a properly formatted command to the manager.

    :param cmd: The command to send (:class:`str`).
    :param expect_reply: If ``True``, wait for the manager to reply
        with an acknowledgement packet.
    :param extra_fields: A dictionary of extra information that's
        merged into the packet body (i.e. not under an extra key).
    """
    packet = {
        ExecutorProtocol.DATA_ID: DATA["id"],
        ExecutorProtocol.COMMAND: cmd,
    }
    packet.update(extra_fields)

    logger.debug("Sending command to listener: {}".format(json.dumps(packet)))

    # TODO what happens here if the push fails? we don't have any realistic recourse,
    # so just let it explode and stop processing
    queue_channel = EXECUTOR_SETTINGS["REDIS_CHANNEL_PAIR"][0]
    try:
        await redis_conn.rpush(queue_channel, json.dumps(packet))
    except Exception:
        logger.error(
            "Error sending command to manager:\n\n{}".format(traceback.format_exc())
        )
        raise

    if not expect_reply:
        return

    for _ in range(_REDIS_RETRIES):
        response = await redis_conn.blpop(QUEUE_RESPONSE_CHANNEL, timeout=1)
        if response:
            break
    else:
        # NOTE: If there's still no response after a few seconds, the system is broken
        # enough that it makes sense to give up; we're isolated here, so if the manager
        # doesn't respond, we can't really do much more than just crash
        raise RuntimeError(
            "No response from the manager after {} retries.".format(_REDIS_RETRIES)
        )

    _, item = response
    result = json.loads(item.decode("utf-8"))[ExecutorProtocol.RESULT]
    assert result in [ExecutorProtocol.RESULT_OK, ExecutorProtocol.RESULT_ERROR]

    if result == ExecutorProtocol.RESULT_OK:
        return True

    return False



"""Search router."""
from rest_framework.routers import DefaultRouter, Route


class SearchRouter(DefaultRouter):
    """Custom router for search endpoints.

    Search endpoints don't follow REST principles and thus don't need
    routes that default router provides.

    """

    routes = [
        Route(
            url=r"^{prefix}{trailing_slash}$",
            mapping={"get": "list", "post": "list_with_post"},
            name="{basename}",
            initkwargs={},
            detail=False,
        )
    ]



# pylint: disable=missing-docstring,invalid-name
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser, Group
from django.contrib.contenttypes.models import ContentType
from django.db.models.query import QuerySet

from guardian.compat import get_user_permission_full_codename
from guardian.exceptions import MixedContentTypeError, WrongAppError
from guardian.shortcuts import assign_perm, get_objects_for_group, remove_perm
from rest_framework import status
from rest_framework.test import APIRequestFactory, force_authenticate

from resolwe.flow.models import Collection, Data, Process
from resolwe.flow.views import StorageViewSet
from resolwe.permissions.shortcuts import (
    get_object_perms,
    get_objects_for_user,
    get_user_group_perms,
)
from resolwe.test import TestCase

factory = APIRequestFactory()


class UserGroupTestCase(TestCase):
    def setUp(self):
        super().setUp()

        self.group1 = Group.objects.create(name="Test group 1")
        self.group2 = Group.objects.create(name="Test group 2")

        self.collection = Collection.objects.create(
            contributor=self.contributor, name="Test collection",
        )

        # This collection is here to make sure that other permissions
        # don't affect tested queries.
        collection2 = Collection.objects.create(
            contributor=self.contributor, name="Test collection 2",
        )
        assign_perm("view_collection", self.contributor, collection2)
        assign_perm("view_collection", self.group1, collection2)

    def test_user(self):
        assign_perm("view_collection", self.contributor, self.collection)
        assign_perm("edit_collection", self.contributor, self.collection)

        user_perms, group_perms = get_user_group_perms(
            self.contributor, self.collection
        )

        self.assertEqual(len(group_perms), 0)
        self.assertCountEqual(user_perms, ["view_collection", "edit_collection"])

    def test_user_in_group(self):
        self.group1.user_set.add(self.contributor)
        assign_perm("view_collection", self.group1, self.collection)
        assign_perm("edit_collection", self.group1, self.collection)

        user_perms, group_perms = get_user_group_perms(
            self.contributor, self.collection
        )
        self.assertEqual(len(group_perms), 1)
        self.assertCountEqual(group_perms[0][2], ["view_collection", "edit_collection"])
        self.assertEqual(len(user_perms), 0)

        assign_perm("view_collection", self.contributor, self.collection)

        user_perms, group_perms = get_user_group_perms(
            self.contributor, self.collection
        )
        self.assertEqual(len(group_perms), 1)
        self.assertCountEqual(group_perms[0][2], ["view_collection", "edit_collection"])
        self.assertEqual(len(user_perms), 1)
        self.assertCountEqual(user_perms, ["view_collection"])

    def test_user_in_multiple_groups(self):
        self.group1.user_set.add(self.contributor)
        self.group2.user_set.add(self.contributor)
        assign_perm("view_collection", self.group1, self.collection)
        assign_perm("edit_collection", self.group1, self.collection)
        assign_perm("view_collection", self.group2, self.collection)

        user_perms, group_perms = get_user_group_perms(
            self.contributor, self.collection
        )
        self.assertEqual(len(group_perms), 2)
        self.assertEqual(group_perms[0][0], self.group1.pk)
        self.assertCountEqual(group_perms[0][2], ["view_collection", "edit_collection"])
        self.assertEqual(group_perms[1][0], self.group2.pk)
        self.assertCountEqual(group_perms[1][2], ["view_collection"])
        self.assertEqual(len(user_perms), 0)

    def test_group(self):
        assign_perm("view_collection", self.group1, self.collection)
        assign_perm("edit_collection", self.group1, self.collection)
        user_perms, group_perms = get_user_group_perms(self.group1, self.collection)
        self.assertEqual(len(group_perms), 1)
        self.assertCountEqual(group_perms[0][2], ["view_collection", "edit_collection"])
        self.assertEqual(len(user_perms), 0)


class ObjectPermsTestCase(TestCase):
    def setUp(self):
        super().setUp()

        self.admin.delete()
        self.user1 = get_user_model().objects.create(username="test_user1")
        self.user2 = get_user_model().objects.create(username="test_user2")
        self.group1 = Group.objects.create(name="Test group 1")
        self.group2 = Group.objects.create(name="Test group 2")
        self.anonymous = AnonymousUser()

        self.collection = Collection.objects.create(
            contributor=self.user1, name="Test collection",
        )

    def _sort_perms(self, perms):
        for elm in perms:
            elm["permissions"] = sorted(elm["permissions"])
        return perms

    def test_all_permissions(self):
        self.group1.user_set.add(self.user1)

        perms = get_object_perms(self.collection)
        self.assertEqual(len(perms), 0)

        assign_perm("view_collection", self.user1, self.collection)
        assign_perm("edit_collection", self.user1, self.collection)
        assign_perm("view_collection", self.user2, self.collection)
        expected_perms = [
            {
                "permissions": ["edit", "view"],
                "type": "user",
                "id": self.user1.pk,
                "name": "test_user1",
                "username": "test_user1",
            },
            {
                "permissions": ["view"],
                "type": "user",
                "id": self.user2.pk,
                "name": "test_user2",
                "username": "test_user2",
            },
        ]
        perms = get_object_perms(self.collection)
        self.assertCountEqual(self._sort_perms(expected_perms), self._sort_perms(perms))

        assign_perm("view_collection", self.group1, self.collection)
        assign_perm("edit_collection", self.group1, self.collection)
        assign_perm("view_collection", self.group2, self.collection)
        expected_perms.extend(
            [
                {
                    "permissions": ["edit", "view"],
                    "type": "group",
                    "id": self.group1.pk,
                    "name": "Test group 1",
                },
                {
                    "permissions": ["view"],
                    "type": "group",
                    "id": self.group2.pk,
                    "name": "Test group 2",
                },
            ]
        )
        perms = get_object_perms(self.collection)
        self.assertCountEqual(self._sort_perms(expected_perms), self._sort_perms(perms))

        assign_perm("view_collection", self.anonymous, self.collection)
        expected_perms.append({"permissions": ["view"], "type": "public"},)
        perms = get_object_perms(self.collection)
        self.assertCountEqual(self._sort_perms(expected_perms), self._sort_perms(perms))

    def test_user_permissions(self):
        self.group1.user_set.add(self.user1)
        assign_perm("view_collection", self.user1, self.collection)
        assign_perm("edit_collection", self.user1, self.collection)
        assign_perm("view_collection", self.user2, self.collection)
        assign_perm("view_collection", self.group1, self.collection)
        assign_perm("edit_collection", self.group1, self.collection)
        assign_perm("view_collection", self.group2, self.collection)

        expected_perms = [
            {
                "permissions": ["edit", "view"],
                "type": "user",
                "id": self.user1.pk,
                "name": "test_user1",
                "username": "test_user1",
            },
            {
                "permissions": ["edit", "view"],
                "type": "group",
                "id": self.group1.pk,
                "name": "Test group 1",
            },
        ]
        perms = get_object_perms(self.collection, self.user1)

        self.assertCountEqual(self._sort_perms(expected_perms), self._sort_perms(perms))

        self.group2.user_set.add(self.user1)
        expected_perms.append(
            {
                "permissions": ["view"],
                "type": "group",
                "id": self.group2.pk,
                "name": "Test group 2",
            },
        )
        perms = get_object_perms(self.collection, self.user1)
        self.assertCountEqual(self._sort_perms(expected_perms), self._sort_perms(perms))

        assign_perm("view_collection", self.anonymous, self.collection)
        expected_perms.append({"permissions": ["view"], "type": "public"},)
        perms = get_object_perms(self.collection, self.user1)
        self.assertCountEqual(self._sort_perms(expected_perms), self._sort_perms(perms))


class StoragePermsTestCase(TestCase):
    def setUp(self):
        super().setUp()

        proc = Process.objects.create(name="Test process", contributor=self.contributor)
        self.data = Data.objects.create(
            name="Test data", contributor=self.contributor, process=proc
        )
        dummy_data = Data.objects.create(
            name="Dummy data", contributor=self.contributor, process=proc
        )

        self.storage1 = self.data.storages.create(
            name="Test storage", json={}, contributor=self.contributor,
        )

        self.storage2 = self.data.storages.create(
            name="Test storage 2", json={}, contributor=self.contributor,
        )

        dummy_data.storages.create(
            name="Dummy storage", json={}, contributor=self.contributor,
        )

        self.user = get_user_model().objects.create(username="test_user")
        self.group = Group.objects.create(name="test_group")

        self.storage_list_viewset = StorageViewSet.as_view(actions={"get": "list",})

        self.storage_detail_viewset = StorageViewSet.as_view(
            actions={"get": "retrieve",}
        )

    def test_detail_permissons(self):
        request = factory.get("/", content_type="application/json")
        force_authenticate(request, self.user)

        resp = self.storage_detail_viewset(request, pk=self.storage1.pk)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)

        assign_perm("view_data", self.user, self.data)
        resp = self.storage_detail_viewset(request, pk=self.storage1.pk)
        self.assertEqual(resp.data["name"], "Test storage")

        remove_perm("view_data", self.user, self.data)
        resp = self.storage_detail_viewset(request, pk=self.storage1.pk)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)

    def test_not_allowed_methods(self):
        self.assertFalse(hasattr(StorageViewSet, "update"))
        self.assertFalse(hasattr(StorageViewSet, "partial_update"))
        self.assertFalse(hasattr(StorageViewSet, "destroy"))
        self.assertFalse(hasattr(StorageViewSet, "create"))


# tests copied from guardina.testapp.tests.test_shortcuts
class GetObjectsForUser(TestCase):
    def setUp(self):
        super().setUp()

        self.group = Group.objects.create(name="group")
        self.ctype = ContentType.objects.create(
            model="bar", app_label="fake-for-guardian-tests"
        )

    def test_superuser(self):
        ctypes = ContentType.objects.all()
        objects = get_objects_for_user(
            self.admin, ["contenttypes.change_contenttype"], ctypes
        )
        self.assertEqual(set(ctypes), set(objects))

    def test_with_superuser_true(self):
        ctypes = ContentType.objects.all()
        objects = get_objects_for_user(
            self.admin, ["contenttypes.change_contenttype"], ctypes, with_superuser=True
        )
        self.assertEqual(set(ctypes), set(objects))

    def test_with_superuser_false(self):
        ctypes = ContentType.objects.all()
        obj1 = ContentType.objects.create(model="foo", app_label="guardian-tests")
        assign_perm("change_contenttype", self.admin, obj1)
        objects = get_objects_for_user(
            self.admin,
            ["contenttypes.change_contenttype"],
            ctypes,
            with_superuser=False,
        )
        self.assertEqual(set([obj1]), set(objects))

    def test_anonymous(self):
        ctypes = ContentType.objects.all()
        objects = get_objects_for_user(
            self.contributor, ["contenttypes.change_contenttype"], ctypes
        )

        obj1 = ContentType.objects.create(model="foo", app_label="guardian-tests")
        assign_perm("change_contenttype", self.contributor, obj1)
        objects = get_objects_for_user(
            self.contributor, ["contenttypes.change_contenttype"], ctypes
        )
        self.assertEqual(set([obj1]), set(objects))

    def test_mixed_perms(self):
        codenames = [
            get_user_permission_full_codename("change"),
            "auth.change_permission",
        ]
        self.assertRaises(
            MixedContentTypeError, get_objects_for_user, self.contributor, codenames
        )

    def test_perms_with_mixed_apps(self):
        codenames = [
            get_user_permission_full_codename("change"),
            "contenttypes.change_contenttype",
        ]
        self.assertRaises(
            MixedContentTypeError, get_objects_for_user, self.contributor, codenames
        )

    def test_mixed_perms_and_klass(self):
        self.assertRaises(
            MixedContentTypeError,
            get_objects_for_user,
            self.contributor,
            ["auth.change_group"],
            get_user_model(),
        )

    def test_no_app_label_nor_klass(self):
        self.assertRaises(
            WrongAppError, get_objects_for_user, self.contributor, ["change_group"]
        )

    def test_empty_perms_sequence(self):
        objects = get_objects_for_user(self.contributor, [], Group.objects.all())
        self.assertEqual(set(objects), set())

    def test_perms_single(self):
        perm = "auth.change_group"
        assign_perm(perm, self.contributor, self.group)
        self.assertEqual(
            set(get_objects_for_user(self.contributor, perm)),
            set(get_objects_for_user(self.contributor, [perm])),
        )

    def test_klass_as_model(self):
        assign_perm("contenttypes.change_contenttype", self.contributor, self.ctype)

        objects = get_objects_for_user(
            self.contributor, ["contenttypes.change_contenttype"], ContentType
        )
        self.assertEqual([obj.name for obj in objects], [self.ctype.name])

    def test_klass_as_manager(self):
        assign_perm("auth.change_group", self.contributor, self.group)
        objects = get_objects_for_user(
            self.contributor, ["auth.change_group"], Group.objects
        )
        self.assertEqual([obj.name for obj in objects], [self.group.name])

    def test_klass_as_queryset(self):
        assign_perm("auth.change_group", self.contributor, self.group)
        objects = get_objects_for_user(
            self.contributor, ["auth.change_group"], Group.objects.all()
        )
        self.assertEqual([obj.name for obj in objects], [self.group.name])

    def test_ensure_returns_queryset(self):
        objects = get_objects_for_user(self.contributor, ["auth.change_group"])
        self.assertTrue(isinstance(objects, QuerySet))

    def test_simple(self):
        group_names = ["group1", "group2", "group3"]
        groups = [Group.objects.create(name=name) for name in group_names]
        for group in groups:
            assign_perm("change_group", self.contributor, group)

        objects = get_objects_for_user(self.contributor, ["auth.change_group"])
        self.assertEqual(len(objects), len(groups))
        self.assertTrue(isinstance(objects, QuerySet))
        self.assertEqual(set(objects), set(groups))

    def test_multi_perms(self):
        group_names = ["group1", "group2", "group3"]
        groups = [Group.objects.create(name=name) for name in group_names]
        for group in groups:
            assign_perm("auth.change_group", self.contributor, group)
        assign_perm("auth.delete_group", self.contributor, groups[1])

        objects = get_objects_for_user(
            self.contributor, ["auth.change_group", "auth.delete_group"]
        )
        self.assertEqual(len(objects), 1)
        self.assertTrue(isinstance(objects, QuerySet))
        self.assertEqual(
            set(objects.values_list("name", flat=True)), set([groups[1].name])
        )

    def test_multi_perms_no_groups(self):
        group_names = ["group1", "group2", "group3"]
        groups = [Group.objects.create(name=name) for name in group_names]
        for group in groups:
            assign_perm("auth.change_group", self.contributor, group)
        assign_perm("auth.delete_group", self.contributor, groups[1])

        objects = get_objects_for_user(
            self.contributor,
            ["auth.change_group", "auth.delete_group"],
            use_groups=False,
        )
        self.assertEqual(len(objects), 1)
        self.assertTrue(isinstance(objects, QuerySet))
        self.assertEqual(
            set(objects.values_list("name", flat=True)), set([groups[1].name])
        )

    def test_any_of_multi_perms(self):
        group_names = ["group1", "group2", "group3"]
        groups = [Group.objects.create(name=name) for name in group_names]
        assign_perm("auth.change_group", self.contributor, groups[0])
        assign_perm("auth.delete_group", self.contributor, groups[2])

        objects = get_objects_for_user(
            self.contributor, ["auth.change_group", "auth.delete_group"], any_perm=True
        )
        self.assertEqual(len(objects), 2)
        self.assertTrue(isinstance(objects, QuerySet))
        self.assertEqual(
            set(objects.values_list("name", flat=True)),
            set([groups[0].name, groups[2].name]),
        )

    def test_groups_perms(self):
        group1 = Group.objects.create(name="group1")
        group2 = Group.objects.create(name="group2")
        group3 = Group.objects.create(name="group3")
        groups = [group1, group2, group3]
        for group in groups:
            self.contributor.groups.add(group)

        # Objects to operate on
        ctypes = list(ContentType.objects.all().order_by("id"))
        assign_perm("auth.change_group", self.contributor)
        assign_perm("change_contenttype", self.contributor, ctypes[0])
        assign_perm("change_contenttype", self.contributor, ctypes[1])
        assign_perm("delete_contenttype", self.contributor, ctypes[1])
        assign_perm("delete_contenttype", self.contributor, ctypes[2])

        assign_perm("change_contenttype", groups[0], ctypes[3])
        assign_perm("change_contenttype", groups[1], ctypes[3])
        assign_perm("change_contenttype", groups[2], ctypes[4])
        assign_perm("delete_contenttype", groups[0], ctypes[0])

        objects = get_objects_for_user(
            self.contributor, ["contenttypes.change_contenttype"]
        )
        self.assertEqual(
            set(objects.values_list("id", flat=True)),
            set(ctypes[i].id for i in [0, 1, 3, 4]),
        )

        objects = get_objects_for_user(
            self.contributor,
            ["contenttypes.change_contenttype", "contenttypes.delete_contenttype"],
        )
        self.assertEqual(
            set(objects.values_list("id", flat=True)), set(ctypes[i].id for i in [0, 1])
        )

        objects = get_objects_for_user(
            self.contributor, ["contenttypes.change_contenttype"]
        )
        self.assertEqual(
            set(objects.values_list("id", flat=True)),
            set(ctypes[i].id for i in [0, 1, 3, 4]),
        )

    def test_has_global_permission_only(self):
        group_names = ["group1", "group2", "group3"]
        for name in group_names:
            Group.objects.create(name=name)

        # global permission to change any group
        perm = "auth.change_group"

        assign_perm(perm, self.contributor)
        objects = get_objects_for_user(self.contributor, perm)
        remove_perm(perm, self.contributor)
        self.assertEqual(set(objects), set(Group.objects.all()))

    def test_has_global_permission_and_object_based_permission(self):
        group_names = ["group1", "group2", "group3"]
        groups = [Group.objects.create(name=name) for name in group_names]
        # global permission to change any group
        perm_global = "auth.change_group"
        perm_obj = "delete_group"
        assign_perm(perm_global, self.contributor)
        assign_perm(perm_obj, self.contributor, groups[0])
        objects = get_objects_for_user(self.contributor, [perm_global, perm_obj])
        remove_perm(perm_global, self.contributor)
        self.assertEqual(
            set(objects.values_list("name", flat=True)), set([groups[0].name])
        )

    def test_has_global_permission_and_object_based_permission_any_perm(self):
        group_names = ["group1", "group2", "group3"]
        groups = [Group.objects.create(name=name) for name in group_names]
        # global permission to change any group
        perm_global = "auth.change_group"
        # object based permission to change only a specific group
        perm_obj = "auth.delete_group"
        assign_perm(perm_global, self.contributor)
        assign_perm(perm_obj, self.contributor, groups[0])
        objects = get_objects_for_user(
            self.contributor,
            [perm_global, perm_obj],
            any_perm=True,
            accept_global_perms=True,
        )
        remove_perm(perm_global, self.contributor)
        self.assertEqual(set(objects), set(Group.objects.all()))

    def test_object_based_permission_without_global_permission(self):
        group_names = ["group1", "group2", "group3"]
        groups = [Group.objects.create(name=name) for name in group_names]
        # global permission to delete any group
        perm_global = "auth.delete_group"
        perm_obj = "auth.delete_group"
        assign_perm(perm_global, self.contributor)
        assign_perm(perm_obj, self.contributor, groups[0])
        objects = get_objects_for_user(
            self.contributor, [perm_obj], accept_global_perms=False
        )
        remove_perm(perm_global, self.contributor)
        self.assertEqual(
            set(objects.values_list("name", flat=True)), set([groups[0].name])
        )

    def test_object_based_permission_with_groups_2perms(self):
        group_names = ["group1", "group2", "group3"]
        groups = [Group.objects.create(name=name) for name in group_names]
        for group in groups:
            self.contributor.groups.add(group)
        # Objects to operate on
        ctypes = list(ContentType.objects.all().order_by("id"))
        assign_perm("contenttypes.change_contenttype", self.contributor)
        assign_perm("change_contenttype", self.contributor, ctypes[0])
        assign_perm("change_contenttype", self.contributor, ctypes[1])
        assign_perm("delete_contenttype", self.contributor, ctypes[1])
        assign_perm("delete_contenttype", self.contributor, ctypes[2])

        assign_perm("change_contenttype", groups[0], ctypes[3])
        assign_perm("change_contenttype", groups[1], ctypes[3])
        assign_perm("change_contenttype", groups[2], ctypes[4])
        assign_perm("delete_contenttype", groups[0], ctypes[0])

        objects = get_objects_for_user(
            self.contributor,
            ["contenttypes.change_contenttype", "contenttypes.delete_contenttype"],
            accept_global_perms=True,
        )
        self.assertEqual(
            set(objects.values_list("id", flat=True)),
            set([ctypes[0].id, ctypes[1].id, ctypes[2].id]),
        )

    def test_object_based_permission_with_groups_3perms(self):

        group_names = ["group1", "group2", "group3"]
        groups = [Group.objects.create(name=name) for name in group_names]
        for group in groups:
            self.contributor.groups.add(group)
        # Objects to operate on
        ctypes = list(ContentType.objects.all().order_by("id"))
        assign_perm("contenttypes.change_contenttype", self.contributor)
        assign_perm("change_contenttype", self.contributor, ctypes[0])
        assign_perm("change_contenttype", self.contributor, ctypes[1])
        assign_perm("delete_contenttype", self.contributor, ctypes[1])
        assign_perm("delete_contenttype", self.contributor, ctypes[2])
        # add_contenttype does not make sense, here just for testing purposes,
        # to also cover one if branch in function.
        assign_perm("add_contenttype", self.contributor, ctypes[1])

        assign_perm("change_contenttype", groups[0], ctypes[3])
        assign_perm("change_contenttype", groups[1], ctypes[3])
        assign_perm("change_contenttype", groups[2], ctypes[4])
        assign_perm("delete_contenttype", groups[0], ctypes[0])
        assign_perm("add_contenttype", groups[0], ctypes[0])

        objects = get_objects_for_user(
            self.contributor,
            [
                "contenttypes.change_contenttype",
                "contenttypes.delete_contenttype",
                "contenttypes.add_contenttype",
            ],
            accept_global_perms=True,
        )
        self.assertEqual(
            set(objects.values_list("id", flat=True)), set([ctypes[0].id, ctypes[1].id])
        )

    def test_exception_different_ctypes(self):
        self.assertRaises(
            MixedContentTypeError,
            get_objects_for_user,
            self.contributor,
            ["auth.change_permission", "auth.change_group"],
        )

    def test_has_any_permissions(self):
        # We use groups as objects.
        group_names = ["group1", "group2", "group3"]
        groups = [Group.objects.create(name=name) for name in group_names]
        for group in groups:
            assign_perm("change_group", self.contributor, group)

        objects = get_objects_for_user(self.contributor, [], Group)
        self.assertEqual(len(objects), len(groups))
        self.assertTrue(isinstance(objects, QuerySet))
        self.assertEqual(set(objects), set(groups))

    def test_short_codenames_with_klass(self):
        assign_perm("contenttypes.change_contenttype", self.contributor, self.ctype)

        objects = get_objects_for_user(
            self.contributor, ["change_contenttype"], ContentType
        )
        self.assertEqual([obj.name for obj in objects], [self.ctype.name])

    def test_has_any_group_permissions(self):
        # We use groups as objects.
        group_names = ["group1", "group2", "group3"]
        groups = [Group.objects.create(name=name) for name in group_names]
        for group in groups:
            assign_perm("change_group", self.group, group)

        objects = get_objects_for_group(self.group, [], Group)
        self.assertEqual(len(objects), len(groups))
        self.assertTrue(isinstance(objects, QuerySet))
        self.assertEqual(set(objects), set(groups))



"""Elastic Search indexes for Collection model."""
import elasticsearch_dsl as dsl

from resolwe.elastic.indices import BaseIndex

from ..models import Collection
from .base import BaseDocument, BaseIndexMixin


class CollectionDocument(BaseDocument):
    """Document for collection search."""

    # Data values extracted from the descriptor.
    descriptor_data = dsl.Text(multi=True)
    tags = dsl.Keyword(multi=True)
    description = dsl.Text(fielddata=True)

    class Index:
        """Meta class for collection search document."""

        name = "collection"


class CollectionIndexMixin:
    """Mixin for indices for collection objects used in ``CollectionDocument``."""

    def extract_descriptor(self, obj):
        """Extract data from the descriptor."""
        descriptor = []

        def flatten(current):
            """Flatten descriptor."""
            if isinstance(current, dict):
                for key in current:
                    flatten(current[key])
            elif isinstance(current, list):
                for val in current:
                    flatten(val)
            elif isinstance(current, (int, bool, float, str)):
                descriptor.append(str(current))

        flatten(obj.descriptor)

        return descriptor

    def get_descriptor_data_value(self, obj):
        """Extract data from the descriptors."""
        return self.extract_descriptor(obj)


class CollectionIndex(BaseIndexMixin, CollectionIndexMixin, BaseIndex):
    """Index for collection objects used in ``CollectionDocument``."""

    queryset = Collection.objects.all().prefetch_related(
        "descriptor_schema", "contributor"
    )
    object_type = Collection
    document_class = CollectionDocument



"""Resolwe collection serializer."""
from resolwe.flow.models import Collection, DescriptorSchema
from resolwe.rest.fields import ProjectableJSONField

from .base import ResolweBaseSerializer
from .descriptor import DescriptorSchemaSerializer
from .fields import DictRelatedField


class CollectionSerializer(ResolweBaseSerializer):
    """Serializer for Collection objects."""

    settings = ProjectableJSONField(required=False)
    descriptor = ProjectableJSONField(required=False)
    descriptor_schema = DictRelatedField(
        queryset=DescriptorSchema.objects.all(),
        serializer=DescriptorSchemaSerializer,
        allow_null=True,
        required=False,
    )

    class Meta:
        """CollectionSerializer Meta options."""

        model = Collection
        read_only_fields = (
            "created",
            "descriptor_dirty",
            "duplicated",
            "id",
            "modified",
        )
        update_protected_fields = ("contributor",)
        fields = (
            read_only_fields
            + update_protected_fields
            + (
                "description",
                "descriptor",
                "descriptor_schema",
                "name",
                "settings",
                "slug",
                "tags",
            )
        )



# pylint: disable=missing-docstring
from rest_framework import mixins, serializers, viewsets

from resolwe.elastic.viewsets import (
    ElasticSearchBaseViewSet,
    ElasticSearchCombinedViewSet,
)

from .elastic_indexes import TestSearchDocument
from .models import TestModel


class TestSerializer(serializers.Serializer):
    name = serializers.CharField()
    num = serializers.IntegerField()
    json = serializers.JSONField(source="json.to_dict")


class TestViewSet(ElasticSearchBaseViewSet):

    document_class = TestSearchDocument

    serializer_class = TestSerializer

    filtering_fields = ("name", "num", "date", "field_process_type", "name_alias")
    filtering_map = {"name_alias": "name"}
    ordering_fields = ("name",)
    ordering = "-name"


class TestEmptyOrderingViewSet(ElasticSearchBaseViewSet):
    document_class = TestSearchDocument
    serializer_class = TestSerializer


class TestCustomFieldFilterViewSet(ElasticSearchBaseViewSet):
    document_class = TestSearchDocument
    serializer_class = TestSerializer
    filtering_fields = ("name",)

    def custom_filter_name(self, value, search):
        """Test for custom filter name, which queries the 'num' field instead of 'name'."""
        return search.query("match", num=value)


class TestModelSerializer(serializers.ModelSerializer):
    class Meta:
        model = TestModel
        fields = ("id", "name", "field_process_type", "number")


class TestModelViewSet(mixins.ListModelMixin, viewsets.GenericViewSet):
    serializer_class = TestModelSerializer
    queryset = TestModel.objects.all()


class TestCombinedViewSet(ElasticSearchCombinedViewSet, TestModelViewSet):
    document_class = TestSearchDocument

    filtering_fields = ("name",)
    ordering_map = {"name": "field_name.raw"}
    ordering_fields = ("name", "num")
